#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
This module implements a model (in the `MVC` sense) for the real data stored
in a `tables.Leaf`.
"""

import logging

import tables

from qtpy import QtCore

import vitables.utils
from vitables.vttables import buffer
from vitables.vttables import filenodebuffer

__docformat__ = 'restructuredtext'

#: The maximum number of rows to be read from the data source.
CHUNK_SIZE = 10000

log = logging.getLogger(__name__)


class LeafModel(QtCore.QAbstractTableModel):
    """
    The model for real data contained in leaves.

    The data is read from data sources (i.e., `HDF5/PyTables` nodes) by
    the model.
    The dataset number of rows is potentially huge but tables are read and
    displayed in chunks.


    :param parent:
        The parent of the model, passed as is in the superclass.
    :attribute leaf:
        the underlying hdf5 data
    :attribute rbuffer:
        Code for chunking and inspecting the undelying data.
    :attribute leaf_numrows:
        the total number of rows in the underlying data
    :attribute numrows:
        The number of rows visible which equals the chunking-size.
    :attribute numcols:
        The total number of columnss visible, equal to those visible.
    :attribute start:
        The zero-based starting index of the chunk within the total rows.

    """

    def __init__(self, leaf, parent=None):
        """Create the model.
        """

        # The model data source (a PyTables/HDF5 leaf) and its access buffer
        self.leaf = leaf
        self.is_filenode = False
        vtapp = vitables.utils.getApp()
        if leaf in vtapp.filenodes_map:
            self.is_filenode = True
            self.rbuffer = filenodebuffer.FilenodeBuffer(leaf)
        else:
            self.rbuffer = buffer.Buffer(leaf)

        self.leaf_numrows = self.rbuffer.total_nrows()
        self.numrows = min(self.leaf_numrows, CHUNK_SIZE)
        self.start = 0

        # The dataset number of columns doesn't use to be large so, we don't
        # need set a maximum as we did with rows. The whole set of columns
        # are displayed
        if isinstance(leaf, tables.Table):
            # Leaf is a PyTables table
            self.numcols = len(leaf.colnames)
        elif isinstance(leaf, tables.EArray):
            self.numcols = 1
        else:
            # Leaf is some kind of PyTables array
            shape = leaf.shape
            if len(shape) > 1:
                # The leaf will be displayed as a bidimensional matrix
                self.numcols = shape[1]
            else:
                # The leaf will be displayed as a column vector
                self.numcols = 1

        #
        # Choose a format for cells
        #

        self.formatContent = vitables.utils.formatArrayContent

        # Time series (if they are found) are formatted transparently
        # via the time_series.py plugin

        if self.is_filenode:
            self.formatContent = vitables.utils.formatStringContent
        elif not isinstance(leaf, tables.Table):
            # Leaf is some kind of PyTables array
            atom_type = leaf.atom.type
            if atom_type == 'object':
                self.formatContent = vitables.utils.formatObjectContent
            elif atom_type in ('vlstring', 'vlunicode'):
                self.formatContent = vitables.utils.formatStringContent
            else:
                self.formatContent = vitables.utils.formatArrayContent
        else:
            self.formatContent = vitables.utils.formatArrayContent

        # Track selected cell
        self.selected_cell = {'index': QtCore.QModelIndex(), 'buffer_start': 0}

        # Populate the model with the first chunk of data
        self.loadData(0, self.numrows)

        super(LeafModel, self).__init__(parent)

    def columnCount(self, index=QtCore.QModelIndex()):
        """The number of columns of the given model index.

        Overridden to return 0 for valid indices because they have no children;
        otherwise return the total number of *columns* exposed by the model.

        :param index:
            the model index being inspected.
        """

        return 0 if index.isValid() else self.numcols

    def rowCount(self, index=QtCore.QModelIndex()):
        """The number of columns for the children of the given index.

        Overridden to return 0 for valid indices because they have no children;
        otherwise return the total number of *rows* exposed by the model.

        :Parameter index: the model index being inspected.
        """

        return 0 if index.isValid() else self.numrows

    def loadData(self, start, length):
        """Load the model with fresh data from the buffer.

        :param start:
            the document row that is the first row of the chunk.
        :param length:
            the buffer size, i.e. the number of rows to be read.

        :return:
            a tuple with tested values for the parameters of the read method
        """

        # Enforce scrolling limits.
        #
        start = max(start, 0)
        stop = min(start + length, self.leaf_numrows)

        # Ensure buffer filled when scrolled beyond bottom.
        #
        actual_start = stop - self.numrows
        start = max(min(actual_start, start), 0)

        self.rbuffer.readBuffer(start, stop)
        self.start = start

    def get_corner_span(self):
        """Must return ``(row_span, col_span)`` tuple for the top-left cell."""
        return 1, 1

    def headerData(self, section, orientation, role):
        """Returns the data for the given role and section in the header
        with the specified orientation.

        This is an overwritten method.

        :Parameters:

        - `section`: the header section being inspected
        - `orientation`: the header orientation (horizontal or vertical)
        - `role`: the role of the header section being inspected
        """

        # The section alignment
        if role == QtCore.Qt.TextAlignmentRole:
            if orientation == QtCore.Qt.Horizontal:
                return QtCore.Qt.AlignLeft | QtCore.Qt.AlignVCenter
            return QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter

        if role != QtCore.Qt.DisplayRole:
            return None

        # Columns-labels
        if orientation == QtCore.Qt.Horizontal:
            # For tables horizontal labels are column names, for arrays
            # the section numbers are used as horizontal labels
            if hasattr(self.leaf, 'description'):
                return str(self.leaf.colnames[section])
            return str(section)

        # Rows-labels
        return str(self.start + section)

    def data(self, index, role=QtCore.Qt.DisplayRole):
        """Returns the data stored under the given role for the item
        referred to by the index.

        This is an overwritten method.

        :Parameters:

        - `index`: the index of a data item
        - `role`: the role being returned
        """
        row, col = index.row(), index.column()

        if not index.isValid() or not (0 <= row < self.numrows):
            return None

        if role == QtCore.Qt.DisplayRole:
            cell = self.cell(row, col)
            return self.formatContent(cell)

        if role == QtCore.Qt.TextAlignmentRole:
            return QtCore.Qt.AlignLeft | QtCore.Qt.AlignTop

        return None

    def cell(self, row, col):
        """
        Returns the contents of a cell.

        :return: none to disable zooming.
        """
        try:
            return self.rbuffer.getCell(row, col)
        except IndexError:
            log.error('IndexError! buffer start: {0} row, column: '
                      '{1}, {2}'.format(self.start, row, col))



#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""The plugins loader module.

For information about the pkg_resources module see
http://setuptools.readthedocs.io/en/latest/pkg_resources.html
"""

__docformat__ = 'restructuredtext'

import pkg_resources
import logging
import traceback

from qtpy import QtGui
from qtpy import QtWidgets

translate = QtWidgets.QApplication.translate

PLUGIN_GROUP = 'vitables.plugins'

log = logging.getLogger(__name__)


def _add_plugin_translator(module_name, plugin_class):
    """Try to load plugin translator."""

    app = QtWidgets.QApplication.instance()
    if hasattr(plugin_class, 'translator'):
        try:
            app.installTranslator(plugin_class.translator)
        except Exception as e:
            log.error('Failed to install {0} plugin '
                      'translator'.format(module_name))
            log.error(e)


def _load_entrypoint(entrypoint):
    """Try to load an entry point."""

    plugin_class = None
    try:
        plugin_class = entrypoint.load()
    except Exception as e:
        log.error('Failed to load plugin: {0}'.format(
            entrypoint.module_name))
        log.error(e)
        log.info(traceback.format_exc())
    return plugin_class


def _create_instance(plugin_class):
    """Try to create an instance of the given plugin class."""

    instance = None
    try:
        instance = plugin_class()
    except Exception as e:
        log.error('Failed to create plugin instance')
        log.error(e)
        log.info(traceback.format_exc())
    return instance


class PluginsLoader(object):
    """Plugins loader class.

    The class loads enabled plugins and their translations and stores
    info about them.

    """

    def __init__(self, enabled_plugins):
        """Assign default values to members.
        """

        # list of UID of enabled plugins, stored in configuration
        self.enabled_plugins = enabled_plugins
        # dictionary that contains all available plugins
        self.all_plugins = {}
        # instances of loaded plugins
        self.loaded_plugins = {}

    def _disable_not_loaded(self):
        self.enabled_plugins = list(self.loaded_plugins.keys())

    def loadAll(self):
        """Find plugins in the system and create instances of enabled ones."""

        for entrypoint in pkg_resources.iter_entry_points(PLUGIN_GROUP):
            plugin_class = _load_entrypoint(entrypoint)
            if plugin_class is None:
                continue
            self.all_plugins[plugin_class.UID] = {
                'UID': plugin_class.UID, 'name': plugin_class.NAME,
                'comment': plugin_class.COMMENT}
            if plugin_class.UID in self.enabled_plugins:
                _add_plugin_translator(entrypoint.module_name,
                                       plugin_class)
                instance = _create_instance(plugin_class)
                if instance is not None:
                    self.loaded_plugins[plugin_class.UID] = instance
        self._disable_not_loaded()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
Here is defined the VTSplash class. It displays the splash screen during the
``ViTables`` startup and shows the advance of the startup sequence.
"""

import time

from qtpy import QtCore
from qtpy import QtGui
from qtpy import QtWidgets


__docformat__ = 'restructuredtext'


class VTSplash(QtWidgets.QSplashScreen):
    """
    The application splash screen.

    :Parameter png: the pixmap image displayed as a splash screen.
    """

    def __init__(self, png):
        """
        Initialize the application.

        Create a splash screen and ties it to a painter which will
        be in charge of displaying the needed messages.
        """

        super(VTSplash, self).__init__(png)
        self.msg = ''

    def drawContents(self, painter):
        """Draw the contents of the splash screen using the given painter.

        This is an overloaded method. It draws contents with the origin
        translated by a certain offset.

        :Parameter painter: the painter used to draw the splash screen
        """

        painter.setPen(QtGui.QColor(QtCore.Qt.white))
        font = painter.font()
        font.setBold(True)
        painter.setFont(font)
        painter.drawText(10, 200, self.msg)

    def drawMessage(self, msg):
        """
        Draw the message text onto the splash screen.

        :Parameter msg: the message to be displayed
        """

        QtWidgets.qApp.processEvents()
        self.msg = msg
        self.showMessage(self.msg)
        time.sleep(0.500)
        self.clearMessage()



import sys
import os.path
import nose.tools as nt

import sip
sip.setapi('QString', 2)
sip.setapi('QVariant', 2)

from qtpy import QtWidgets
from qtpy import QtCore
import qtpy.QtTest as qtest

import vitables.start as vtstart

from vitables.vtapp import VTApp
from vitables.preferences import vtconfig

TABLES_PATH = os.path.join('examples', 'tables')


def get_index(model, node_name, start_index=None):
    """Return index of the node with the given name or None."""
    start_index = start_index if start_index else QtCore.QModelIndex()
    for c in range(model.columnCount(start_index)):
        for r in range(model.rowCount(start_index)):
            index = model.index(r, c, start_index)
            if index.data(QtCore.Qt.DisplayRole) == node_name:
                return index
            child_search_result = get_index(model, node_name, index)
            if child_search_result:
                return child_search_result
    return None


def assert_path(test, path):
    """Activate and check that node at given path exist.

    If the last node is a leaf then subwindow will be created.
    """
    if not path:
        return
    index = get_index(test.model, path[0])
    nt.assert_false(index is None)
    test.view.activateNode(index)
    assert_path(test, path[1:])


def get_leaf_model_with_assert(test, filepath, leaf_path):
    """Open view for given leaf and return its model.

    filepath is a valid system path, node_path is a slash separated leaf
    path inside file without preceding slash.
    """
    leaf_path = leaf_path.split('/')
    _, filename = os.path.split(filepath)
    test.vtapp.fileOpen(filepath)
    assert_path(test, [filename] + leaf_path)
    return test.vtgui.workspace.subWindowList()[0].leaf_model


class TestTableOpening:
    """Open files and read all cell values."""

    TEST_NODES = {
        'examples/arrays/array_samples.h5': [
            'array_b', 'array_f', 'array_f3D', 'array_int16', 'array_int8',
            'group1/array_B', 'group2/array_h', 'group3/array_l',
            'group4/array_f', 'group5/array_d', 'scalar_array',
        ],
        'examples/arrays/carray_sample.h5': [
            'carray',
        ],
        'examples/arrays/earray_samples.h5': [
            'array_b', 'array_c', 'array_char', 'array_e',
        ],
        'examples/arrays/vlarray_samples.h5': [
            'vlarray1', 'vlarray11', 'vlarray2', 'vlarray3', 'vlarray5',
            'vlarray6', 'vlarray8', 'vlarray10', 'vlarray9',
            # 'vlarray7',
        ],
        'examples/misc/external_file.h5': [
            'a1',
        ],
        # modified on open !!!
        # 'examples/misc/fnode.h5': [
        #     'fnode_test',
        # ],
        # modified on open !!!
        # 'examples/misc/genericHDF5.h5': [
        #     'A note', 'arrays/2D float array', 'arrays/2D int array',
        #     'arrays/3D int array',
        #     'arrays/Vdata table: PerBlockMetadataCommon',
        #     'arrays/external',
        #     # 'images/iceberg_palette',
        #     # 'images/Iceberg',
        #     # 'images/landcover.umd.199906.jpg',
        #     # 'images/pixel interlace',
        #     # 'images/plane interlace',
        # ],
        'examples/misc/links_examples.h5': [
            'arrays/a1', 'links/ht1', 'tables/t1',
        ],
        # modified on open !!!
        # 'examples/misc/nonalphanum.h5': [
        #     'array_1', 'array_f', 'array_f_Numeric', 'array_s',
        # ],
        'examples/misc/objecttree.h5': [
            'array1', 'group1/array2', 'group1/table1', 'group2/table2',
        ],
        'examples/misc/szip_compressor.h5': [
            # 'datasetF32',
        ],
        'examples/tables/nested_samples.h5': [
            'table',
        ],
        'examples/tables/table_samples.h5': [
            'columns/TDC', 'columns/name', 'columns/pressure',
            'detector/recarray', 'detector/recarray2', 'detector/table',
            'newgroup/table', 'table1',
        ],
        'examples/timeseries/carray_ts.h5': [
            'test_carray_1', 'test_carray_2',
        ],
        'examples/timeseries/pandas_test1.hdf5': [
            # 'one_column_ts/_i_table/index/abounds',
            # 'one_column_ts/_i_table/index/bounds',
            # 'one_column_ts/_i_table/index/indices',
            # 'one_column_ts/_i_table/index/indicesLR',
            # 'one_column_ts/_i_table/index/mbounds',
            # 'one_column_ts/_i_table/index/mranges',
            # 'one_column_ts/_i_table/index/ranges',
            # 'one_column_ts/_i_table/index/sorted',
            # 'one_column_ts/_i_table/index/sortedLR',
            # 'one_column_ts/_i_table/index/zbounds',
            # 'one_column_ts/table',
        ],
        'examples/timeseries/pandas_test2.hdf5': [
            # 'intc/_i_table/index/abounds', 'intc/_i_table/index/bounds',
            # 'intc/_i_table/index/indices', 'intc/_i_table/index/indicesLR',
            # 'intc/_i_table/index/mbounds', 'intc/_i_table/index/mranges',
            # 'intc/_i_table/index/ranges', 'intc/_i_table/index/sorted',
            # 'intc/_i_table/index/sortedLR', 'intc/_i_table/index/zbounds',
            # 'intc/table',
        ],
        'examples/timeseries/pandas_test3.hdf5': [
            # 'df/axis0', 'df/axis1', 'df/block0_items', 'df/block0_values',
            # 'df_table/_i_table/index/abounds',
            # 'df_table/_i_table/index/bounds',
            # 'df_table/_i_table/index/indices',
            # 'df_table/_i_table/index/indicesLR',
            # 'df_table/_i_table/index/mbounds',
            # 'df_table/_i_table/index/mranges',
            # 'df_table/_i_table/index/ranges',
            # 'df_table/_i_table/index/sorted',
            # 'df_table/_i_table/index/sortedLR',
            # 'df_table/_i_table/index/zbounds',
            'df_table/table',
        ],
        # modified on open !!!
        # 'examples/timeseries/scikits_test1.hdf5': [
        #     'examples/Example_1',
        # ],
        # modified on open !!!
        # 'examples/timeseries/scikits_test2.hdf5': [
        #     'examples/Example_2',
        # ],
        # modified on open !!!
        # 'examples/timeseries/scikits_test3.hdf5': [
        #     'examples/Example_3',
        # ],
        'examples/timeseries/table_ts.h5': [
            'Particles/TParticle',
        ],
    }

    @classmethod
    def setup_class(cls):
        """Create app and store shortcuts to the application objects."""
        cls.app = QtWidgets.QApplication(sys.argv)
        cls.vtapp = VTApp(keep_splash=False)
        cls.vtgui = cls.vtapp.gui
        cls.model = cls.vtgui.dbs_tree_model
        cls.view = cls.vtgui.dbs_tree_view

    def test_opening_files(self):
        """Generate tests for given data files."""
        for filepath, nodes in self.TEST_NODES.items():
            for node in nodes:
                yield self.check_node_open, filepath, node

    def check_node_open(self, filepath, nodepath):
        """Open file get access to a node and read all cells."""
        try:
            leaf_model = get_leaf_model_with_assert(self, filepath, nodepath)
            for row in range(leaf_model.rowCount()):
                for column in range(leaf_model.columnCount()):
                    cell_index = leaf_model.index(row, column,
                                                  QtCore.QModelIndex())
                    cell_data = leaf_model.data(cell_index)
        finally:
            self.vtapp.fileClose()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

#
#       This script is based on a set of scripts by Francesc Alted.

"""Several simple EArrays."""

import tables
import numpy

fileh = tables.open_file('earray_samples.h5', mode='w')

root = fileh.root
a = tables.StringAtom(itemsize=8)
# Use ``a`` as the object type for the enlargeable array.
array_c = fileh.create_earray(root, 'array_c', a, (0,), "Chars")
array_c.append(numpy.array(['a'*2, 'b'*4], dtype='S8'))
array_c.append(numpy.array(['a'*6, 'b'*8, 'c'*10], dtype='S8'))

# Create an string atom
a = tables.StringAtom(itemsize=1)
# Use it as a type for the enlargeable array
hdfarray = fileh.create_earray(root, 'array_char', a, (0,), "Character array")
hdfarray.append(numpy.array(['a', 'b', 'c']))
# The next is legal:
hdfarray.append(numpy.array(['c', 'b', 'c', 'd']))
# but these are not:
#hdfarray.append(array([['c', 'b'], ['c', 'd']]))
#hdfarray.append(array([[1,2,3],[3,2,1]], dtype=uint8).reshape(2,1,3))

# Create an atom
a = tables.UInt16Atom()
hdfarray = fileh.create_earray(root, 'array_e', a, (2, 0, 3), 
    "Unsigned short array")

# Create an enlargeable array
a = tables.UInt8Atom()
hdfarray = fileh.create_earray(root, 'array_b', a, (2, 0, 3), 
    "Unsigned byte array", tables.Filters(complevel = 1))

# Append an array to this table
hdfarray.append(numpy.array([[1, 2, 3], [3, 2, 1]], 
    dtype=numpy.uint8).reshape(2, 1, 3))
hdfarray.append(numpy.array([[1, 2, 3], [3, 2, 1], [2, 4, 6], [6, 4, 2]], 
    dtype=numpy.uint8).reshape(2,2,3)*2)
# The next should give a type error:
#hdfarray.append(array([[1,0,1],[0,0,1]], dtype=Bool).reshape(2,1,3))

# Close the file.
fileh.close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""CArray with time fields.
"""

import os
import time

import tables

output_dir = '../timeseries'
try:
    os.mkdir(output_dir)
except OSError:
    pass

# Open a new empty HDF5 file
hdf5_name = "carray_ts.h5"
filepath_hdf5 = os.path.join(output_dir, hdf5_name)
h5file = tables.open_file(filepath_hdf5, mode="w",
title='Example CArray with time fields')

# Create a CArray and fill it
root = h5file.root
shape = (300, 2)
atom = tables.Time32Atom()
filters = tables.Filters(complevel=5, complib='zlib')
hdfarray = h5file.create_carray(root, 'test_carray_1', atom, shape, 
    "Signed short array")
now = time.time()
seconds_by_day = 1*24*60*60
for index in range(0, 600, 2):
    seconds = now - seconds_by_day * index 
    hdfarray[index/2, 0] = seconds
    hdfarray[299-index/2, 1] = seconds

# Create other CArray and fill it
shape = (300,)
atom = tables.Time32Atom(shape=(2,))
filters = tables.Filters(complevel=5, complib='zlib')
hdfarray = h5file.create_carray(root, 'test_carray_2', atom, shape, 
    "Signed short array")
now = time.time()
seconds_by_day = 1*24*60*60
for index in range(0, 600, 2):
    seconds = now - seconds_by_day * index 
    hdfarray[index/2] = [seconds, seconds]

h5file.close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

#
#       This script is based on a set of scripts by Francesc Alted.

"""Several simple Tables."""

import sys

import numpy
import tables

class Particle(tables.IsDescription):
    """Description of a table record.
    """
    name        = tables.StringCol(16, pos=1)   # 16-character String
    lati        = tables.Int32Col(pos=2)        # integer
    longi       = tables.Int32Col(pos=3)        # integer
    pressure    = tables.Float32Col(pos=4)      # float  (single-precision)
    temperature = tables.Float64Col(pos=5)      # double (double-precision)

# Open a file in "w"rite mode
fileh = tables.open_file("table_samples.h5", mode = "w")

root = fileh.root
# Create a new group
group = fileh.create_group(root, "newgroup")

# Create a new table in newgroup group
table = fileh.create_table(group, 'table', Particle, "A table", 
    tables.Filters(1))
particle = table.row

# Fill the table with 10 particles
for i in range(10):
    # First, assign the values to the Particle record
    particle['name']  = 'Particle: %6d' % (i)
    particle['lati'] = i
    particle['longi'] = 10 - i
    particle['pressure'] = float(i*i)
    particle['temperature'] = float(i**2)
    # This injects the row values.
    particle.append()

# We need to flush the buffers in table in order to get an
# accurate number of records on it.
table.flush()

# Add a couple of user attrs
table.attrs.user_attr1 = 1.023
table.attrs.user_attr2 = "This is the second user attr"

# Append several rows in only one call
table.append([("Particle:     10", 10, 0, 10*10, 10**2),
              ("Particle:     11", 11, -1, 11*11, 11**2),
              ("Particle:     12", 12, -2, 12*12, 12**2)])
table.flush()

class Particle2(tables.IsDescription):
    """Description of a table record.
    """
    name        = tables.StringCol(16, pos=1)   # 16-character String
    lati        = tables.ComplexCol(itemsize=16, pos=2)
    longi       = tables.ComplexCol(itemsize=8, pos=3)
    vector      = tables.ComplexCol(itemsize=8, shape=(2,), pos=4)
    matrix2D    = tables.ComplexCol(itemsize=16, shape=(2, 2), pos=5)

# Open a file in "w"rite mode
table1 = fileh.create_table(root, 'table1', Particle2, "A table")
# Append several rows in only one call
table1.append([("Particle:     10", 10j, 0, (10*9+1j, 1), [[10**2j, 11*3]]*2),
              ("Particle:     11", 11j, -1, (11*10+2j, 2), [[11**2j, 10*3]]*2),
              ("Particle:     12", 12j, -2, (12*11+3j, 3), [[12**2j, 9*3]]*2),
              ("Particle:     13", 13j, -3, (13*11+4j, 4), [[13**2j, 8*3]]*2),
              ("Particle:     14", 14j, -4, (14*11+5j, 5), [[14**2j, 7*3]]*2)])
table1.flush()

######
class Particle3(tables.IsDescription):
    ADCcount    = tables.Int16Col()              # signed short integer
    TDCcount    = tables.UInt8Col()              # unsigned byte
    grid_i      = tables.Int32Col()              # integer
    grid_j      = tables.Int32Col()              # integer
    idnumber    = tables.Int64Col()              # signed long long
    name        = tables.StringCol(16, dflt="")  # 16-character String
    pressure    = tables.Float32Col(shape=2)     # float  (single-precision)
    temperature = tables.Float64Col()            # double (double-precision)

Particle4 = {
    # You can also use any of the atom factories, i.e. the one which
    # accepts a PyTables type.
    "ADCcount"    : tables.Col.from_type("int16"),    # signed short integer
    "TDCcount"    : tables.Col.from_type("uint8"),    # unsigned byte
    "grid_i"      : tables.Col.from_type("int32"),    # integer
    "grid_j"      : tables.Col.from_type("int32"),    # integer
    "idnumber"    : tables.Col.from_type("int64"),    # signed long long
    "name"        : tables.Col.from_kind("string", 16),  # 16-character String
    "pressure"    : tables.Col.from_type("float32", (2,)), # float  (single-precision)
    "temperature" : tables.Col.from_type("float64"),  # double (double-precision)
}

# Create a new group under "/" (root)
group = fileh.create_group("/", 'detector')

# Create one table on it
#table = h5file.create_table(group, 'table', Particle, "Title example")
# You can choose creating a Table from a description dictionary if you wish
table2 = fileh.create_table(group, 'table', Particle4, "Title example")

# Create a shortcut to the table record object
particle = table2.row

# Fill the table with 10 particles
for i in range(10):
    # First, assign the values to the Particle record
    particle['name']  = 'Particle: %6d' % (i)
    particle['TDCcount'] = i % 256
    particle['ADCcount'] = (i * 256) % (1 << 16)
    particle['grid_i'] = i
    particle['grid_j'] = 10 - i
    particle['pressure'] = [float(i*i), float(i*2)]
    particle['temperature'] = float(i**2)
    particle['idnumber'] = i * (2 ** 34)  # This exceeds integer range
    # This injects the Record values.
    particle.append()

# Flush the buffers for table
table2.flush()

# Create a new group to hold new arrays
gcolumns = fileh.create_group("/", "columns")
pressure = [ p['pressure'] for p in table2.iterrows() ]
# Create an array with this info under '/columns' having a 'list' flavor
fileh.create_array(gcolumns, 'pressure', pressure, 
                   "Pressure column")

# Do the same with TDCcount, but with a numpy object
TDC = [ p['TDCcount'] for p in table2.iterrows() ]
fileh.create_array('/columns', 'TDC', numpy.array(TDC), "TDCcount column")

# Do the same with name column
names = [ p['name'] for p in table2.iterrows() ]
fileh.create_array('/columns', 'name', names, "Name column")

# Save a recarray object under detector
r = numpy.rec.array("a"*300, formats='f4,3i4,a5,i2', shape=3)
recarrt = fileh.create_table("/detector", 'recarray', r, "RecArray example")
r2 = r[0:3:2]
# Change the byteorder property
recarrt = fileh.create_table("/detector", 'recarray2', r2,
                             "Non-contiguous recarray")

# Finally, append some new records to table
table3 = fileh.root.detector.table

# Append 5 new particles to table (yes, tables can be enlarged!)
particle = table3.row
for i in range(10, 15):
    # First, assign the values to the Particle record
    particle['name']  = 'Particle: %6d' % (i)
    particle['TDCcount'] = i % 256
    particle['ADCcount'] = (i * 256) % (1 << 16)
    particle['grid_i'] = i
    particle['grid_j'] = 10 - i
    particle['pressure'] = [float(i*i), float(i*2)]
    particle['temperature'] = float(i**2)
    particle['idnumber'] = i * (2 ** 34)  # This exceeds integer range
    # This injects the Row values.
    particle.append()

# Flush this table
table3.flush()

# Finally, close the file
fileh.close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
Here is defined the HelpBrowserGUI class.

This module creates the main window for the documentation browser.
"""

__docformat__ = 'restructuredtext'

import re

from qtpy import QtCore
from qtpy import QtGui
from qtpy import QtWidgets

import vitables.utils
from vitables.vtsite import DOCDIR


translate = QtWidgets.QApplication.translate


class HelpBrowserGUI(QtWidgets.QMainWindow) :
    """
    The main window of the documentation browser.

    :Parameters:

    - `browser`: an instance of the docs browser controller
      :meth:`vitables.docbrowser.helpbrowser.HelpBrowser`
    - `parent`: the parent widget.
    """

    def __init__(self, browser, parent=None) :
        """
        Initializes the browser.
        """

        super(HelpBrowserGUI, self).__init__(parent)

        self.setIconSize(QtCore.QSize(22, 22))
        self.setWindowTitle(translate('HelpBrowserGUI',
            'Documentation browser', 'The window title'))
        self.icons = vitables.utils.getHBIcons()
        self.setWindowIcon(self.icons['vitables_wm'])

        self.browser = browser

        # The browser widget
        self.text_browser = QtWidgets.QTextBrowser()
        self.text_browser.setSearchPaths([DOCDIR])
        self.setCentralWidget(self.text_browser)
        self.text_browser.setAcceptRichText(True)
        self.text_browser.setReadOnly(1)

        # The popup menus
        self.actions = self.setupActions()
        self.initPopups()
        self.connectSignals()
        self.setupHistoryCombo()
        self.statusBar().showMessage(translate('HelpBrowserGUI', 'Ready...',
                                    'Status bar startup message'))


    def setupActions(self):
        """Provide actions to the menubar and the toolbars.
        """

        actions = {}

        actions['exitBrowser'] = QtWidgets.QAction(
            translate('HelpBrowserGUI', 'E&xit', 'File --> Exit'), self,
            shortcut=QtGui.QKeySequence.Quit,
            triggered=self.browser.exitBrowser,
            icon=self.icons['application-exit'],
            statusTip=translate('HelpBrowserGUI', 'Close Help Browser',
                    'Status bar text for the File --> Exit action'))

        actions['zoomIn'] = QtWidgets.QAction(
            translate('HelpBrowserGUI', 'Zoom &in', 'View --> Zoom in'), self,
            shortcut=QtGui.QKeySequence.ZoomIn,
            triggered=self.browser.zoomIn,
            icon=self.icons['zoom-in'],
            statusTip=translate('HelpBrowserGUI', 'Increases the font size',
                'Status bar text for the View --> Zoom in action'))

        actions['zoomOut'] = QtWidgets.QAction(
            translate('HelpBrowserGUI', 'Zoom &out', 'View --> Zoom out'),
            self,
            shortcut=QtGui.QKeySequence.ZoomOut,
            triggered=self.browser.zoomOut,
            icon=self.icons['zoom-out'],
            statusTip=translate('HelpBrowserGUI', 'Decreases the font size',
                'Status bar text for the View --> Zoom out action'))

        actions['goHome'] = QtWidgets.QAction(
            translate('HelpBrowserGUI', '&Home', 'Go --> Home'), self,
            shortcut=QtGui.QKeySequence.UnknownKey,
            triggered=self.text_browser.home,
            icon=self.icons['go-first-view'],
            statusTip=translate('HelpBrowserGUI',
                'Go to the first visited page',
                'Status bar text for the  Go --> Home action'))

        actions['goBackward'] = QtWidgets.QAction(
            translate('HelpBrowserGUI', '&Backward', ' Go --> Backward'),
            self,
            shortcut=QtGui.QKeySequence.Back,
            triggered=self.text_browser.backward,
            icon=self.icons['go-previous-view'],
            statusTip=translate('HelpBrowserGUI', 'Go to previous page',
                'Status bar text for the  Go --> Backward action'))

        actions['goForward'] = QtWidgets.QAction(
            translate('HelpBrowserGUI', '&Forward', ' Go --> Forward'), self,
            shortcut=QtGui.QKeySequence.Forward,
            triggered=self.text_browser.forward,
            icon=self.icons['go-next-view'],
            statusTip=translate('HelpBrowserGUI', 'Go to next page',
                'Status bar text for the  Go --> Forward action'))

        actions['goReload'] = QtWidgets.QAction(
            translate('HelpBrowserGUI', '&Reload', 'Go --> Reload'), self,
            shortcut=QtGui.QKeySequence.Refresh,
            triggered=self.text_browser.reload,
            icon=self.icons['view-refresh'],
            statusTip=translate('HelpBrowserGUI', 'Reload the current page',
                'Status bar text for the  Go --> Reload action'))

        actions['bookmarksAdd'] = QtWidgets.QAction(
            translate('HelpBrowserGUI', '&Add bookmark',
                'Bookmarks --> Add bookmark'),
            self,
            shortcut=QtGui.QKeySequence('Ctrl+Alt+N'),
            triggered=self.browser.addBookmark,
            icon=self.icons['bookmark_add'],
            statusTip=translate('HelpBrowserGUI', 'Bookmark the current page',
                'Status bar text for Bookmarks --> Add bookmark action'))

        actions['bookmarksEdit'] = QtWidgets.QAction(
            translate('HelpBrowserGUI', '&Edit bookmarks...',
                'Bookmarks --> Edit bookmarks'),
            self,
            shortcut=QtGui.QKeySequence('Ctrl+Alt+E'),
            triggered=self.browser.editBookmarks,
            icon=self.icons['bookmarks'],
            statusTip=translate('HelpBrowserGUI', 'Edit bookmarks',
                'Status bar text for Bookmarks --> Edit bookmarks action'))

        actions['bookmarksClear'] = QtWidgets.QAction(
            translate('HelpBrowserGUI', '&Clear All',
                'Bookmarks --> Clear bookmark'),
            self,
            shortcut=QtGui.QKeySequence('Ctrl+Alt+C'),
            triggered=self.browser.clearBookmarks,
            statusTip=translate('HelpBrowserGUI',
                'Clear all existing bookmarks',
                'Status bar text for Bookmarks --> Add bookmark action'))

        actions['about'] = QtWidgets.QAction(
            translate('HelpBrowserGUI', '&About HelpBrowser',
                'Help --> About HelpBrowser'),
            self,
            shortcut=QtGui.QKeySequence.UnknownKey,
            triggered=self.browser.aboutBrowser,
            statusTip=translate('HelpBrowserGUI', 'About HelpBrowser',
                'Status bar text for Help --> About HelpBrowser action'))

        actions['aboutQt'] = QtWidgets.QAction(
            translate('HelpBrowserGUI', 'About &Qt', 'Help --> About Qt'),
            self,
            shortcut=QtGui.QKeySequence.UnknownKey,
            triggered=self.browser.aboutQt,
            statusTip=translate('HelpBrowserGUI', 'About Qt',
                'Status bar text for the Help --> About Qt action'))

        actions['clearSession'] = QtWidgets.QAction(
            translate('HelpBrowserGUI', 'Clear history', ''), self,
            shortcut=QtGui.QKeySequence.UnknownKey,
            triggered=self.browser.clearHistory,
            icon=self.icons['edit-clear-history'],
            statusTip=translate('HelpBrowserGUI',
                'Clear the content of the history combobox', ''))

        return actions


    def initPopups(self) :
        """
        Setup the menubar and the toolbar of the main window.

        The menubar contains the menus `File`, `Go`, `Bookmarks` and `Help`.
        The toolbar contains the buttons: `home`, `backward`, `forward`,
        `combobox` and `clear history`.
        """

        # Create the File menu and add actions/submenus/separators to it
        file_menu = self.menuBar().addMenu(
            translate('HelpBrowserGUI', "&File", 'The File menu entry'))
        file_actions = ['exitBrowser']
        vitables.utils.addActions(file_menu, file_actions, self.actions)


        # Create the View menu and toolbar
        view_menu = self.menuBar().addMenu(
            translate('HelpBrowserGUI', "&View", 'The View menu entry'))
        view_toolbar = QtWidgets.QToolBar(
            translate('HelpBrowserGUI', 'View operations', 'Toolbar title'),
            self)
        self.addToolBar(view_toolbar)
        view_actions = ['zoomIn', 'zoomOut']
        vitables.utils.addActions(view_menu, view_actions, self.actions)
        vitables.utils.addActions(view_toolbar, view_actions, self.actions)

        # Create the Go menu and toolbar
        go_menu = self.menuBar().addMenu(translate('HelpBrowserGUI', "&Go",
            'The Go menu entry'))
        go_toolbar = QtWidgets.QToolBar(
            translate('HelpBrowserGUI', 'Go operations', 'Toolbar title'),
            self)
        self.addToolBar(go_toolbar)
        go_actions = ['goHome', 'goBackward', 'goForward', 'goReload']
        vitables.utils.addActions(go_menu, go_actions, self.actions)
        vitables.utils.addActions(go_toolbar, go_actions, self.actions)


        # Create the Bookmarks menu and toolbar
        self.bookmarks_menu = self.menuBar().addMenu(
            translate('HelpBrowserGUI', "&Bookmarks", 'Bookmarks menu entry'))
        bookmarks_toolbar = QtWidgets.QToolBar(
            translate('HelpBrowserGUI', 'Bookmarks operations',
            'Toolbar title'), self)
        self.addToolBar(bookmarks_toolbar)
        bookmark_actions = ['bookmarksAdd', 'bookmarksEdit', 'bookmarksClear',
                            None]
        vitables.utils.addActions(self.bookmarks_menu, bookmark_actions,
            self.actions)
        vitables.utils.addActions(bookmarks_toolbar, bookmark_actions[:2],
            self.actions)

        # Create the Help menu and add actions/submenus/separators to it
        help_menu = self.menuBar().addMenu(
            translate('HelpBrowserGUI', "&Help", 'The Help menu entry'))
        help_actions = ['about', 'aboutQt']
        vitables.utils.addActions(help_menu, help_actions, self.actions)

        # Create the History toolbar
        history_toolbar = QtWidgets.QToolBar(
            translate('HelpBrowserGUI', 'History operations', 'Toolbar title'),
            self)
        self.addToolBar(history_toolbar)
        history_actions = ['clearSession']
        vitables.utils.addActions(history_toolbar, history_actions,
            self.actions)
        go_selector = QtWidgets.QLabel(
            translate('HelpBrowserGUI', 'Go: ', 'Text of the Go: label'),
            history_toolbar)
        history_toolbar.addWidget(go_selector)
        self.combo_history = QtWidgets.QComboBox(history_toolbar)
        self.combo_history.setSizeAdjustPolicy(
            QtWidgets.QComboBox.AdjustToContents)
        history_toolbar.addWidget(self.combo_history)


    def connectSignals(self):
        """
        Connect signals to slots.

        Signals coming from GUI are connected to slots in the docs browser
        controller, :meth:`vitables.docbrowser.helpbrowser.HelpBrowser`.
        """

        self.combo_history.activated[str].connect(self.browser.displaySrc)

        # This is the most subtle connection. It encompasses source
        # changes coming from anywhere, including slots (home, backward
        # and forward), menus (Go and Bookmarks), clicked links and
        # programatic changes (setSource calls).
        self.text_browser.sourceChanged.connect(self.browser.updateHistory)

        self.text_browser.backwardAvailable.connect(
            self.browser.updateBackward)

        self.text_browser.forwardAvailable.connect(
            self.browser.updateForward)

        # The Bookmarks menu is special case due to its dynamic nature.
        # The menu content vary every time a bookmark is added/deleted
        # In order to track changes and keep it updated, the menu is reloaded
        # every time it is about to be displayed.
        self.bookmarks_menu.aboutToShow.connect(self.updateRecentSubmenu)


    def updateRecentSubmenu(self):
        """Update the content of the Bookmarks menu."""

        # Clear the current bookmarks from the Bookmarks menu
        for action in self.bookmarks_menu.actions():
            if re.search("^(\s?\d)", action.text()):
                self.bookmarks_menu.removeAction(action)
        # and refresh it
        index = 0
        for filepath in self.browser.bookmarks:
            index += 1
            action = QtWidgets.QAction('{0:>2}. {1}'.format(index, filepath),
                                    self.bookmarks_menu)
            action.setData(filepath)
            self.bookmarks_menu.addAction(action)
            action.triggered.connect(self.browser.displaySrc)


    def setupHistoryCombo(self):
        """
        Initializes history combobox.
        """

        # Setup combobox
        self.combo_history.setEditable(0)
        self.combo_history.setWhatsThis(translate('HelpBrowserGUI',
            """<qt>
            <h3>Page selector</h3>Select the page you want to visit.
            </qt>""",
            'WhatsThis text for the combobox of visited pages')
            )
        for item in self.browser.history :
            self.combo_history.addItem(item)


    def closeEvent(self, event) :
        """
        Reimplements the event handler for `QCloseEvent` events.

        Before the close event is accepted we need to do some stuff. This can
        be done in two different ways: via event filters or reimplementing the
        event handler. We have chosen the second possibility.

        :Parameter event: the event being handled
        """

        # When the help browser window is closed via File --> Exit
        # the exitBrowser is called and its history is saved.
        # But if we close the window with the
        # close button then history is not saved at all.
        # We fix this misbehavior by overwriting this method.
        self.browser.exitBrowser()
        QtWidgets.QMainWindow.closeEvent(self, event)



#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""This is the launcher script for the ViTables application."""

import sys
import locale
import argparse
import os.path
import logging
import traceback

import qtpy.QtCore as qtcore
from qtpy import QtWidgets

from vitables.vtapp import VTApp
from vitables.preferences import vtconfig

__docformat__ = 'restructuredtext'

# Map number of -v's on command line to logging error level.
_VERBOSITY_LOGLEVEL_DICT = {0: logging.ERROR, 1: logging.WARNING,
                            2: logging.INFO, 3: logging.DEBUG}
# Default log format used by logger.
_FILE_LOG_FORMAT = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
# Folder with vitables translations.
_I18N_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'i18n')


_uncaught_exception_logger = logging.getLogger('vitables')


def _uncaught_exception_hook(type_, value, tb):
    _uncaught_exception_logger.error(''.join(traceback.format_tb(tb))
                                     + str(value))
    sys.__excepthook__(type_, value, tb)


sys.excepthook = _uncaught_exception_hook


def _check_versions():
    """Check that tables are at least version 3.0"""
    import tables
    if tables.__version__ < '3.0':
        sys.exit('FATAL: PyTables version 3.0 or above is required, '
                 'installed version is {}'.format(tables.__version__))


def _set_credentials(app):
    """Specify the organization's Internet domain.

    When the Internet domain is set, it is used on Mac OS X instead of
    the organization name, since Mac OS X applications conventionally
    use Internet domains to identify themselves

    """
    app.setOrganizationDomain('vitables.org')
    app.setOrganizationName('ViTables')
    app.setApplicationName('ViTables')
    app.setApplicationVersion(vtconfig.getVersion())


def _set_locale(app):
    """Set locale and load translation if available.

    Localize the application using the system locale numpy seems to
    have problems with decimal separator in some locales (catalan,
    german...) so C locale is always used for numbers.

    """
    locale.setlocale(locale.LC_ALL, '')
    locale.setlocale(locale.LC_NUMERIC, 'C')

    locale_name = qtcore.QLocale.system().name()
    translator = qtcore.QTranslator()
    if translator.load('vitables_' + locale_name, _I18N_PATH):
        app.installTranslator(translator)
    return translator


def _parse_command_line():
    """Create parser and parse command line."""
    # Parse the command line optional arguments
    parser = argparse.ArgumentParser(usage='%(prog)s [option]... [h5file]...')
    h5files_group = parser.add_argument_group('h5files')
    logging_group = parser.add_argument_group('logging')
    # Options for the default group
    parser.add_argument(
        '--version', action='version',
        version='%(prog)s {}'.format(vtconfig.getVersion()))
    # Options for opening files
    h5files_group.add_argument(
        '-m', '--mode', choices=['r', 'a'], metavar='MODE',
        help='mode access for a database. Can be r(ead) or a(ppend)')
    h5files_group.add_argument('-d', '--dblist',
                               help=('a file with the list of HDF5 '
                                     'filepaths to be open'))
    # Logging options
    logging_group.add_argument('-l', '--log-file', help='log file path')
    logging_group.add_argument('-v', '--verbose', action='count', default=0,
                               help='log verbosity level')
    # Allow an optional list of input filepaths
    parser.add_argument('h5file', nargs='*')
    # Set sensible option defaults
    parser.set_defaults(mode='a', dblist='', h5file=[])
    # parse and process arguments
    args = parser.parse_args()
    if args.dblist:
        # Other options and positional arguments are silently ignored
        args.mode = ''
        args.h5file = []
    return args


def _setup_logger(args):
    """Setup logger output format, level and output file.

    Stderr logger is added to handle error that raise before the gui
    is launched. It better be removed before event loop starts.

    """
    logger = logging.getLogger()
    file_formatter = logging.Formatter(_FILE_LOG_FORMAT)
    temporary_stderr_handler = logging.StreamHandler()
    temporary_stderr_handler.setFormatter(file_formatter)
    logger.addHandler(temporary_stderr_handler)
    if args.log_file is not None:
        try:
            log_filename = os.path.expandvars(
                os.path.expanduser(args.log_file))
            file_handler = logging.FileHandler(log_filename)
            file_handler.setFormatter(file_formatter)
            logger.addHandler(file_handler)
        except Exception as e:
            logger.error('Failed to open log file')
            logger.error(e)
    if args.verbose in _VERBOSITY_LOGLEVEL_DICT:
        logger.setLevel(_VERBOSITY_LOGLEVEL_DICT[args.verbose])
    else:
        logger.setLevel(logging.ERROR)
        logger.error('Invalid verbosity level: {}, error level '
                     'set to ERROR'.format(args.verbose))
    return logger, temporary_stderr_handler


def gui():
    """The application launcher.

    First of all, translators are loaded. Then the GUI is shown and
    the events loop is started.

    """
    _check_versions()
    app = QtWidgets.QApplication(sys.argv)
    _set_credentials(app)
    translator = _set_locale(app)  # must not be destroyed before app quits
    args = _parse_command_line()
    logger, console_log_handler = _setup_logger(args)
    vtapp = VTApp(mode=args.mode, dblist=args.dblist, h5files=args.h5file)
    vtapp.gui.show()
    logger.removeHandler(console_log_handler)
    app.exec_()


if __name__ == '__main__':
    gui()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
This minimal module makes a QScrollBar for browsing huge datasets.

The use of this scrollbar is very tricky. It will be painted over the scrollbar
of views bound to huge datasets so the view will look as usual but will have 2
scrollbars. The hidden scrollbar is not useful for browsing huge datasets (it
is tightly bound to the dimensions (number of rows) of the view widget. But the
visible scrollbar is independent of the dimensions of the view widget and can
be used for customising the view in a way that makes it useful for browsing
datasets with a larger number of rows than that provided by the view widget.
"""

__docformat__ = 'restructuredtext'

from qtpy.QtCore import Qt
from qtpy.QtCore import QEvent
from qtpy.QtWidgets import QScrollBar


class ScrollBar(QScrollBar):
    """
    A specialised scrollbar for views of huge datasets.

    :Parameter scrollbar: the scrollbar being hidden
    """

    def __init__(self, view):
        """Replace a vertical scrollbar with other one.

        After replacing, the ancestor widgets of `scrollbar` looks
        exactly the same, but the visible scrollbar is not currently
        useful for navigating the data displayed in the ancestor widgets
        because it is not tied to that data in anyway.
        """

        self.view = view
        # Cheat the user hidding a scrollbar and displaying other one
        # that looks exactly the same
        parent = view.vscrollbar.parent()
        super(ScrollBar, self).__init__(parent)
        view.vscrollbar.setVisible(False)
        parent.layout().addWidget(self)
        self.setOrientation(Qt.Vertical)
        self.setObjectName('tricky_vscrollbar')

    def event(self, e):
        """Filter wheel events and send them to the table viewport. """
        if (e.type() == QEvent.Wheel):
            self.view.wheelEvent(e)
            return True
        return QScrollBar.event(self, e)

    def setMaxValue(self, max_value):
        """Ensure range of scrollbar is a signed 32bit integer."""
        max_value = min(2 ** 31 - 1, max_value)
        self.setMaximum(max_value)

        return max_value



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
Display recursively the content of a given cell of a view.
"""

__docformat__ = 'restructuredtext'

from qtpy import QtCore
from qtpy import QtGui
from qtpy import QtWidgets

import vitables.utils

def getArrayDimensions(shape):
    """
    Get the dimensions of the grid where the cell will be zoomed.

    The zoomed cell contains a `numpy` array and will be displayed
    in a table with the same shape than it.

    :Parameter shape: the cell shape

    :Returns: a tuple (rows, columns)
    """

    # Numpy scalars are special a case
    if shape == ():
        nrows = ncols = 1
    # 1-D arrays are displayed as a 1-column vector with nrows
    elif len(shape) == 1:
        nrows = shape[0]
        ncols = 1
    # N-D arrays are displayed as a matrix (nrowsXncols) which
    # elements are (N-2)-D arrays
    else:
        nrows = shape[0]
        ncols = shape[1]

    return (nrows, ncols)

class ZoomCell(QtWidgets.QMdiSubWindow):
    """
    Display an array/table cell on its own view (table widget).

    When a leaf is displayed in a view, is quite usual that the content
    of some cells is not fully visible because it doesn't fit into
    the cell. To alleviate this problem this class provides a way to,
    recursively, display the content of any cell on its own view.
    The cell content depends on the kind of leaf and the shape of its
    atom. Cells of `Table` views and `E/C/Array` views can be:

    - a ``numpy`` scalar. `Atom` shape is ()
    - a ``numpy`` array. `Atom` shape is not ()

    In addition, cells of `VLArray` views can be:

    - a serialized `Python` object. `Atom` kind is `object`, shape is ()
    - a `Python` string. `Atom` kind is `vlstring`, shape is ()

    Finally, cells of `Table` views also can be:

    - a `numpy.void` object when the cell corresponds to nested field of the
        record

    :Parameters:

        - `data`: the value stored in the cell being zoomed
        - `title`: the base string for the zoomed view title
        - `workspace`: the parent of the zoomed view
        - `leaf`: a LeafNode instance
    """

    def __init__(self, data, title, workspace, leaf):
        """
        Creates a zoom view for a given cell.

        The passed cell is an element of a given dataset. It is always
        a (potentially nested) `numpy` array. See cell accessor methods
        in the `Buffer` class for details.
        """

        self.data = data
        self.title = title
        self.workspace = workspace
        self.data_shape = self.hasShape()
        self.field_names = []

        # Create and customise the widget that will display the zoomed cell
        # The pindex attribute is required to keep working the code for
        # synchronising workspace and tree of databases view.
        # The leaf attribute is required to keep working the code for
        # cleaning the workspace when a file is closed.
        # The WA_DeleteOnClose flag makes that when the widget is
        # closed either programatically (see VTAPP.windowClose)
        # or by the user (clicking the close button in the titlebar)
        # the widget is hidden AND destroyed --> the workspace
        # updates automatically its list of open windows --> the
        # Windows menu content is automatically updated
        super(ZoomCell, self).__init__(workspace)
        self.pindex = None
        self.dbt_leaf = leaf
        self.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        # The internal widget
        self.grid = QtWidgets.QTableWidget()
        self.setWidget(self.grid)
        # Configure the titlebar
        self.setWindowTitle(self.title)
        icons_dictionary = vitables.utils.getIcons()
        self.setWindowIcon(icons_dictionary['zoom-in'])

        # Decide how the cell content will be formatted. Content can be:
        # - a numpy array
        # - either a string or a unicode string
        # - other Python object
        if self.data_shape:
            self.formatContent = vitables.utils.formatArrayContent
        elif isinstance(self.data, str):
            self.formatContent = vitables.utils.formatStringContent
        else:
            self.formatContent = vitables.utils.formatObjectContent

        # Setup grid dimensions
        (nrows, ncols) = self.getGridDimensions()
        self.grid.setColumnCount(ncols)
        self.grid.setRowCount(nrows)

        # Setup grid editing
        self.grid.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)

        # Setup grid horizontal header
        if self.field_names:
            for section in range(0, ncols):
                item = QtWidgets.QTableWidgetItem()
                item.setText(self.field_names[section])
                self.grid.setHorizontalHeaderItem(section, item)
        else:
            for section in range(0, ncols):
                item = QtWidgets.QTableWidgetItem()
                item.setText('{0}'.format(section))
                self.grid.setHorizontalHeaderItem(section, item)

        # Setup grid vertical header
        for section in range(0,nrows):
            item = QtWidgets.QTableWidgetItem()
            item.setText('{0}'.format(section))
            self.grid.setVerticalHeaderItem(section, item)

        # Fill the grid
        if self.field_names:
            self.zoomTable()
        else:
            self.zoomArray()

        self.show()

        rmode = QtWidgets.QHeaderView.Stretch
        if self.grid.columnCount() == 1:
            self.grid.horizontalHeader().setSectionResizeMode(rmode)
        if self.grid.rowCount() == 1:
            self.grid.verticalHeader().setSectionResizeMode(rmode)

        # Connect signals to slots
        self.grid.cellDoubleClicked.connect(self.zoomView)


    def hasShape(self):
        """Find out if the zoomed cell has a shape attribute."""
        return hasattr(self.data, 'shape')


    def getGridDimensions(self):
        """
        Get the dimensions of the grid where the zoomed cell will be displayed.

        :Returns: a tuple (rows, columns)
        """

        if self.data_shape:
            # The cell contains a numpy object
            shape = self.data.shape
            dtype = self.data.dtype
            if dtype.fields is None:
                # Arrays with nested fields come here
                return getArrayDimensions(shape)
            else:
                # Table nested fields come here
                return self.getNestedFieldDimensions()
        else:
            # The cell contains a Python object
            return self.getPyObjectDimensions()


    def getPyObjectDimensions(self):
        """
        Get the dimensions of the grid where the cell will be zoomed.

        The zoomed cell contains a `Python` object.

        :Returns: a tuple (rows, columns)
        """

        if isinstance(self.data, list) or isinstance(self.data, tuple):
            return (len(self.data), 1)
        else:
            return (1, 1)


    def getNestedFieldDimensions(self):
        """
        Get the dimensions of the grid where the cell will be zoomed.

        The zoomed cell contains a nested field (a field made of inner
        fields, than can be nested or not) and will be displayed in a
        table with only one row and one column per (top - 1) level field.

        The dtype.descr attribute describes the inner fields. A field
        description is a tuple with one of the following formats:

        - (field_name, format)  field is not nested, shape=()
        - (field_name, format, shape)   field is not nested, shape!=()
        - (field_name, descr)   field is nested

        :Returns: a tuple (rows, columns)
        """

        self.field_names = [item[0] for item in self.data.dtype.descr]
        ncols = len(self.field_names)
        nrows = 1

        return (nrows, ncols)


    def zoomTable(self):
        """Fill the zoom view with the content of the clicked nested field."""

        for column in range(0, self.grid.columnCount()):
            content = self.data[self.field_names[column]]
            text = self.formatContent(content)
            item = QtWidgets.QTableWidgetItem(text)
            self.grid.setItem(0, column, item)


    def zoomArray(self):
        """Fill the zoom view with the content of the clicked cell."""

        num_rows = self.grid.rowCount()
        num_cols = self.grid.columnCount()
        # Numpy scalars are displayed in a 1x1 grid
        if num_rows == num_cols == 1:
            content = self.data
            text = self.formatContent(content)
            item = QtWidgets.QTableWidgetItem(text)
            self.grid.setItem(0, 0, item)
        # 1-D arrays
        elif num_cols == 1:
            for row in range(0, num_rows):
                content = self.data[row]
                text = self.formatContent(content)
                item = QtWidgets.QTableWidgetItem(text)
                self.grid.setItem(row, 0, item)
        # N-D arrays
        else:
            for row in range(0, num_rows):
                for column in range(0, num_cols):
                    content = self.data[row][column]
                    text = self.formatContent(content)
                    item = QtWidgets.QTableWidgetItem(text)
                    self.grid.setItem(row, column, item)


    def zoomView(self, row, col):
        """Makes the content of the clicked cell fully visible.

        :Parameters:

        - `row`: the row of the clicked cell
        - `col`: the column of the clicked cell
        """

        # Check if the zoom has to be done
        if self.data_shape:
            if not (self.data.shape !=() or self.field_names):
                return
        elif not (isinstance(self.data, list) or isinstance(self.data, tuple)):
            return

        # Get data
        if self.data_shape:
            # Arrays and table nested fields
            if self.field_names:
                cell = self.data[self.field_names[col]]
            elif len(self.data.shape) > 1:
                cell = self.data[row][col]
            elif len(self.data.shape) == 1:
                cell = self.data[row]
        else:
            # Python lists and tuples
            cell = self.data[row]

        # Get caption
        if self.field_names:
            caption = '{0}: {1}[{2}]'.format(self.title,
                self.field_names[col], row)
        else:
            caption = '{0}: ({1}, {2})'.format(self.title, row, col)
        ZoomCell(cell, caption, self.workspace, self.dbt_leaf)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""Storing time series created with scikits.timeseries module in PyTables.
Example 2.

Notes:
 -The dates from the yahoo quotes module get returned as integers, which happen
  to correspond to the integer representation of 'DAILY' frequency dates in the
  scikits.timeseries module.
 -`fill_missing_dates` will insert masked values for any missing data points.
  Note that you could plot the series without doing this, but it would cause
  missing values to be linearly interpolated rather than left empty in the plot
"""

import os
import datetime

from matplotlib.finance import quotes_historical_yahoo
import scikits.timeseries as ts
import scikits.timeseries.lib.tstables as tstab

startdate = datetime.date(2002, 1, 5)
enddate = datetime.date(2003, 12, 1)

# retrieve data from yahoo.
# Data format is [(d, open, close, high, low, volume), ...] where d is
# a floating point representation of the number of days since 01-01-01 UTC
quotes = quotes_historical_yahoo('INTC', startdate, enddate)

# Create a DateArray of daily dates and convert it to business day frequency
dates = ts.date_array([q[0] for q in quotes], freq='DAILY').asfreq('BUSINESS')

opens = [q[1] for q in quotes]

# opens: the data portion of the timeserie
# dates: the date portion of the timeserie
raw_series = ts.time_series(opens, dates)
test_series = raw_series
#test_series = ts.fill_missing_dates(raw_series, fill_value=-1)

# Write to a PyTables file
output_dir = '../timeseries'
try:
    os.mkdir(output_dir)
except OSError:
    pass

hdf5_name = 'scikits_test2.hdf5'
filepath_hdf5 = os.path.join(output_dir, hdf5_name)
h5file = tstab.open_file(filepath_hdf5, mode="w",
title='Example table with csikits time series')
group_doc = h5file.create_group("/", 'examples', 'Test Data')
table = h5file.createTimeSeriesTable(group_doc, 'Example_2', test_series)
h5file.close()




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
This module executes `tables.Table` queries at low level.

It collects information from the `New Query` dialog, processes it and then
executes the query.
"""

__docformat__ = 'restructuredtext'

import tables
import numpy

from qtpy import QtCore
from qtpy import QtGui
from qtpy import QtWidgets

import vitables.utils

translate = QtWidgets.QApplication.translate


class Query(QtCore.QObject):
    """Class implementing a tables.Table query.

    Ideally queries SHOULD be sequentially executed in a secondary thread for
    keeping the GUI responsive (i.e. not frozen) while a query (a potentially
    long-running operation) is taking place. Unfortunately this goal has
    proven difficult to achieve because `PyTables` is not thread-safe. So one
    more task to the TO DO list :-)

    The query is implemented in a clever way that doesn't interfer with the
    lazy population of the tree of databases view: the query results table is
    stored under a hidden group of the temporary database until the query
    finishes. Then it is moved to the root node and becomes visible to the
    world. This way, while the query results table is partially filled it
    is not seen by the lazy population algorithm so it is not added to the
    tree of databases view and neither the user nor ViTables will try to read
    it. So no problems can occur trying to read a partially filled table.

    :Parameters:

    - `table_uid`: UID of the tables.Table instance being queried
    - `table`: the table being queried
    - `qdescr`: dictionary description of the query
    """


    query_completed = QtCore.Signal(bool, str, name="queryCompleted")


    def __init__(self, tmp_h5file, table_uid, table, qdescr):
        """Initialises the query."""

        super(Query, self).__init__()

        self.completed = False
        self.tmp_h5file = tmp_h5file
        self.table_uid = table_uid
        self.table = table
        self.qdescr = qdescr


    def run(self):
        """
        Query a table and add a the result to the temporary database.
        """

        self.queryTable()
        self.query_completed.emit(self.completed, self.table_uid)


    def flushTable(self, ftable):
        """Flush the filtered table and setup some user attributes.

        :Parameters:

        - `ftable`: the filtered table being flushed
        """

        ftable.flush()
        # Set some user attributes that define this filtered table
        asi = ftable.attrs
        asi.query_path = self.qdescr['src_filepath']
        asi.query_table = self.qdescr['src_path']
        asi.query_condition = self.qdescr['title']


    def queryWithIndex(self, src_dict):
        """Do the query (`PyTables` level).
        """

        # The query range is made of numpy scalars with dtype int64
        (start, stop, step) = self.qdescr['rows_range']
        chunk_size = 10000
        div = int((stop - start) // chunk_size)

        # Create the destination table: its first column will contain
        # the indices of the rows selected in the source table so a new
        # description dictionary is needed. Int64 values are necessary
        # to keep full 64-bit indices
        ft_dict = {self.qdescr['indices_field_name']: tables.Int64Col(pos=-1)}
        ft_dict.update(src_dict)
        f_table = self.tmp_h5file.create_table(
            '/_p_query_results',
            self.qdescr['ft_name'],
            ft_dict,
            self.qdescr['title'])

        # Get the array of rows that fulfill the condition
        # Selection is done in several steps. It saves a *huge*
        # amount of memory when querying large tables
        for i in numpy.arange(0, div+1):
            QtWidgets.qApp.processEvents()
            lstart = start + chunk_size*i
            if lstart > stop:
                lstart = stop
            lstop = lstart + chunk_size
            if lstop > stop:
                lstop = stop
            coordinates = self.table.get_where_list(
                self.qdescr['condition'],
                self.qdescr['condvars'],
                start=lstart, stop=lstop, step=step)
            selection = self.table.read_coordinates(coordinates)
            if selection.shape == (0, ):
                continue

            coord_dtype = numpy.dtype(
                [(str(self.qdescr['indices_field_name']), '<i8')])
            new_dtype = numpy.dtype(
                coord_dtype.descr + selection.dtype.descr)

            new_buffer = numpy.empty(selection.shape, dtype=new_dtype)
            for field in selection.dtype.fields:
                new_buffer[field] = selection[field]
            new_buffer[str(self.qdescr['indices_field_name'])] = \
                coordinates
            f_table.append(new_buffer)
            self.flushTable(f_table)

        # Move the intermediate table to its final destination
        self.tmp_h5file.move_node(
            '/_p_query_results/' + self.qdescr['ft_name'],
            '/', newname=self.qdescr['ft_name'],
            overwrite=True)
        self.completed = True


    def queryWithNoIndex(self, src_dict):
        """Do the query (`PyTables` level).
        """

        # The query range is made of numpy scalars with dtype int64
        (start, stop, step) = self.qdescr['rows_range']
        chunk_size = 10000
        div = int((stop - start) // chunk_size)

        # Create the destination table
        f_table = self.tmp_h5file.create_table(
            '/_p_query_results',
            self.qdescr['ft_name'],
            src_dict,
            self.qdescr['title'])

        # Get the array of rows that fulfill the condition
        # Selection is done in several steps. It saves a *huge*
        # amount of memory when querying large tables
        for i in numpy.arange(0, div+1):
            QtWidgets.qApp.processEvents()
            lstart = start + chunk_size*i
            if lstart > stop:
                lstart = stop
            lstop = lstart + chunk_size
            if lstop > stop:
                lstop = stop
            selection = self.table.read_where(
                self.qdescr['condition'],
                self.qdescr['condvars'],
                start=lstart, stop=lstop, step=step)
            f_table.append(selection)
            self.flushTable(f_table)

        # Move the intermediate table to its final destination
        self.tmp_h5file.move_node(
            '/_p_query_results/' + self.qdescr['ft_name'],
            '/', newname=self.qdescr['ft_name'],
            overwrite=True)
        self.completed = True


    def queryTable(self):
        """Do the query (`PyTables` level).
        """

        try:
            src_dict = self.table.description._v_colobjects
            # Add an `indexes` column to the result table
            if self.qdescr['indices_field_name']:
                self.queryWithIndex(src_dict)
            # Do no add an `indexes` column to the result table
            else:
                self.queryWithNoIndex(src_dict)
        except KeyError:
            vitables.utils.formatExceptionInfo()
            self.tmp_h5file.remove_node(
                '/_p_query_results/' + self.qdescr['ft_name'])
        else:
            self.tmp_h5file.flush()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

#
#       This script is based on a script by Francesc Alted.

"""A simple CArray."""

import numpy
import tables

fileName = 'carray_sample.h5'
shape = (200, 300)
atom = tables.UInt8Atom()
filters = tables.Filters(complevel=5, complib='zlib')

h5f = tables.open_file(fileName, 'w')
ca = h5f.create_carray(h5f.root, 'carray', atom, shape, filters=filters)
ca[10:60, 20:70] = numpy.ones((50, 50))
h5f.close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
This module defines a view for the model bound to a `tables.Leaf` node.

This view is used to display the real data stored in a `tables.Leaf` node in a
tabular way:

    - scalar arrays are displayed in a 1x1 table.
    - 1D arrays are displayed in a Mx1 table.
    - KD arrays are displayed in a MxN table.
    - tables are displayed in a MxN table with one field per column

"""

__docformat__ = 'restructuredtext'

from qtpy import QtCore
from qtpy import QtGui
from qtpy import QtWidgets

import vitables.vttables.leaf_delegate as leaf_delegate
import vitables.vttables.scrollbar as scrollbar


_aiv = QtWidgets.QAbstractItemView


class LeafView(QtWidgets.QTableView):
    """
    A view for real data contained in leaves.

    This is a customised view intended to deal with huge datasets.

    :Parameters:

    - `tmodel`: the leaf model to be tied to this view
    - `parent`: the parent of this widget
    """

    def __init__(self, tmodel, parent=None):
        """Create the view.
        """

        super(LeafView, self).__init__(parent)
        self.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        self.tmodel = tmodel  # This is a MUST
        self.leaf_numrows = leaf_numrows = self.tmodel.leaf_numrows
        self.selection_model = self.selectionModel()
        self.setSelectionMode(_aiv.SingleSelection)
        self.setSelectionBehavior(_aiv.SelectItems)

        # Setup the actual vertical scrollbar
        self.setVerticalScrollMode(_aiv.ScrollPerItem)
        self.vscrollbar = self.verticalScrollBar()

        self.setModel(tmodel)

        # For potentially huge datasets use a customised scrollbar
        if leaf_numrows > tmodel.numrows:
            self.setItemDelegate(leaf_delegate.LeafDelegate())
            self.rbuffer_fault = False
            self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOn)
            self.tricky_vscrollbar = scrollbar.ScrollBar(self)
            self.max_value = self.tricky_vscrollbar.setMaxValue(
                self.leaf_numrows)
            self.tricky_vscrollbar.setMinimum(0)
            self.interval_size = self.mapSlider2Leaf()

        # Setup the vertical header width
        self.vheader = QtWidgets.QHeaderView(QtCore.Qt.Vertical)
        self.setVerticalHeader(self.vheader)
        font = self.vheader.font()
        font.setBold(True)
        fmetrics = QtGui.QFontMetrics(font)
        max_width = fmetrics.width(" {0} ".format(str(leaf_numrows)))
        self.vheader.setMinimumWidth(max_width)
        self.vheader.setSectionsClickable(True)

        # Setup the headers' resize mode
        rmode = QtWidgets.QHeaderView.Stretch
        if tmodel.columnCount() == 1:
            self.horizontalHeader().setSectionResizeMode(rmode)
        if tmodel.rowCount() == 1:
            self.vheader.setSectionResizeMode(rmode)

        # Setup the text elide mode
        self.setTextElideMode(QtCore.Qt.ElideRight)

        # Connect signals to slots
        if leaf_numrows > tmodel.numrows:
            self.tricky_vscrollbar.actionTriggered.connect(
                self.navigateWithMouse)

        ## Instead of invoking updateView().
        self.setSpan(0, 0, *tmodel.get_corner_span())

    def mapSlider2Leaf(self):
        """Setup the interval size.

        Get the number of rows we move up/down on the dataset every time the
        slider's value moves up/down by 1 unit. The interval size is given by
        the formula::

            int(self.leaf_numrows/self.max_value)

        Note that the larger is the number of rows the worse is this
        approach. In the worst case we would have an interval size of
        (2**64 - 1)/(2**31 - 1) = 2**33. Nevertheless the approach is
        quite good for number of rows about 2**33 (eight thousand million
        rows). In this case the interval size is about 4.
        """

        # If the slider range equals to the number of rows of the dataset
        # then there is a 1:1 mapping between range values and dataset
        # rows and row equals to value
        interval_size = 1
        if self.max_value < self.leaf_numrows:
            interval_size = round(self.leaf_numrows / self.max_value)
        return interval_size

    def syncView(self):
        """Update the tricky scrollbar value after a data navigation.

        This method every time we navigate on the table (except when we press
        the Home/End keys). Unless we are at top/bottom of the dataset the
        update is done using the first visible row as a reference.
        """

        offset = self.tmodel.start + 1
        fv_label = self.vheader.logicalIndexAt(0) + offset
        lv_label = self.vheader.logicalIndexAt(
            self.vheader.viewport().height() - 1) + offset
        if lv_label == self.leaf_numrows:
            self.tricky_vscrollbar.setValue(self.max_value)
        elif fv_label == 1:
            self.tricky_vscrollbar.setValue(0)
        else:
            value = round(fv_label / self.interval_size)
            self.tricky_vscrollbar.setValue(value)

    def updateView(self):
        """Update the view contents after a buffer fault.
        """
        tmodel = self.tmodel

        self.vheader.headerDataChanged(
            QtCore.Qt.Vertical, 0, tmodel.numrows - 1)
        top_left = tmodel.index(0, 0)
        bottom_right = tmodel.index(tmodel.numrows - 1,
                                    tmodel.numcols - 1)
        self.dataChanged(top_left, bottom_right)
        self.setSpan(0, 0, *tmodel.get_corner_span())

    def navigateWithMouse(self, slider_action):
        """Navigate the view with the mouse.

        On a regular table (with the scrollbar connected to the table view)
        this slot is called after the `action` has set the slider position
        but before the display has been updated (see documentation of the
        QAbstractSlider.actionTriggered signal in the Qt4 docs) so in this
        method we can safely do any action before that display update happens.

        In our case the slot is connected to the tricky scrollbar and its main
        functionality is to refresh model data if needed i.e, to detect buffer
        fault conditions and update the data display accordingly. If a buffer
        fault doesn't occur then the slider action is passed to the hidden
        scrollbar (which is connected to the table view) so that it can update
        the data display.

        :Parameter slider_action: the triggered slider action i.e., a member of
            the QAbstractSlider.SliderAction enum
        """

        # The QAbstractSlider.SliderAction enum values used in this method
        # QtGui.QAbstractSlider.SliderSingleStepAdd -> 1
        # QtGui.QAbstractSlider.SliderSingleStepSub -> 2
        # QtGui.QAbstractSlider.SliderPageStepAdd -> 3
        # QtGui.QAbstractSlider.SliderPageStepSub -> 4
        # QtGui.QAbstractSlider.SliderMove -> 7
        actions = {
            1: self.addSingleStep,
            2: self.subSingleStep,
            3: self.addPageStep,
            4: self.subPageStep,
            7: self.dragSlider
        }
        if slider_action not in actions.keys():
            return
        # Navigate the data dealing with buffer faults
        actions[slider_action]()

        # Eventually synchronize the position of the visible scrollbar
        # with the displayed data using the first visible cell as
        # reference
        self.syncView()

    def mouseNavInfo(self, direction):
        """Gives information about model, vertical header and viewport.

        This is a helper method used by methods that browse the data via mouse.

        :Parameter direction: the data browsing direction (upwards/downwards)
        """

        model = self.tmodel
        vh = self.vheader

        # About the table
        table_rows = model.numrows
        buffer_start = model.start

        # The viewport BEFORE navigating the data
        if (direction == 'u'):
            row = vh.visualIndexAt(0)
        elif (direction == 'd'):
            row = vh.visualIndexAt(self.viewport().height())
        page_step = self.vscrollbar.pageStep()

        return (model, vh, table_rows, buffer_start, row, page_step)

    def addSingleStep(self):
        """Setup data for moving towards the last section line by line.
        """

        model, vh, table_rows, buffer_start, last_vp_row, page_step = \
            self.mouseNavInfo('d')
        # If we are at the last row of the buffer but not at the last
        # row of the dataset we still can go downwards so we have to
        # read the next contiguous buffer
        if (last_vp_row + 1 == table_rows) and \
                (buffer_start + table_rows < self.leaf_numrows):
            # Buffer fault. The new buffer starts just after the current
            # first row of the viewport.
            new_start = buffer_start + last_vp_row - page_step + 1
            model.loadData(new_start, table_rows)
            self.updateView()
            self.scrollTo(
                model.index(new_start - model.start, 0),
                _aiv.PositionAtTop)
        else:
            self.vscrollbar.triggerAction(1)

    def addPageStep(self):
        """Setup data for moving towards the last section page by page.
        """

        model, vh, table_rows, buffer_start, last_vp_row, page_step = \
            self.mouseNavInfo('d')
        # If we are at the last page of the buffer but not at the last
        # row of the dataset we still can go downwards so we have to
        # read the next contiguous buffer
        if (last_vp_row + page_step + 1 > table_rows) and \
                (buffer_start + table_rows < self.leaf_numrows):
            # Buffer fault. The new buffer starts at the current last
            # row of the viewport.
            new_start = buffer_start + last_vp_row
            model.loadData(new_start, table_rows)
            self.updateView()
            self.scrollTo(
                model.index(new_start - model.start, 0),
                _aiv.PositionAtTop)
        else:
            self.vscrollbar.triggerAction(3)

    def subSingleStep(self):
        """Setup data for moving towards the first section line by line.
        """

        model, vh, table_rows, buffer_start, first_vp_row, page_step = \
            self.mouseNavInfo('u')
        # If we are at the first row of the buffer but not at the first
        # row of the dataset we still can go upwards so we have to
        # read the previous contiguous buffer
        if (first_vp_row == 0) and (buffer_start > 0):
            # Buffer fault. The new buffer ends just before the current
            # last row of the viewport.
            model.loadData(
                buffer_start + page_step - table_rows, table_rows)
            self.scrollTo(
                model.index(buffer_start + page_step - model.start - 1, 0),
                _aiv.PositionAtBottom)
            self.updateView()
        else:
            self.vscrollbar.triggerAction(2)

    def subPageStep(self):
        """Setup data for moving towards the first section page by page.
        """

        model, vh, table_rows, buffer_start, first_vp_row, page_step = \
            self.mouseNavInfo('u')
        # If we are at the first page of the buffer but not at the first
        # row of the dataset we still can go upwards so we have to
        # read the previous contiguous buffer
        if (first_vp_row < page_step + 1) and (buffer_start > 0):
            # Buffer fault. The new buffer ends just at the current
            # first row of the viewport.
            model.loadData(
                buffer_start + first_vp_row - table_rows + 1,
                table_rows)
            self.updateView()
            self.scrollTo(
                model.index(
                    buffer_start + first_vp_row - model.start, 0),
                _aiv.PositionAtBottom)
        else:
            self.vscrollbar.triggerAction(4)

    def dragSlider(self):
        """Move the slider by dragging it.

        When navigating large datasets we must beware that the number of
        rows of the dataset (int64) is greater than the number of
        values in the range of values (int32) of the scrollbar. It means
        that there are rows that cannot be reached with the scrollbar.

        Note:: QScrollBar.sliderPosition and QScrollBar.value not always
        return the same value. When we reach the top of the dataset:

        - wheeling: value() returns 0, sliderPosition() returns a
            negative number
        - dragging: value() returns a number greater than 0, sliderPosition()
            returns 0
        """

        model = self.tmodel
        table_rows = model.numrows
        value = self.tricky_vscrollbar.sliderPosition()
        if value < 0:
            value = 0
            row = 0
        elif value >= self.max_value:
            value = self.max_value
            row = self.leaf_numrows - 1
        else:
            row = self.interval_size * value

        # top buffer fault condition
        if row < model.start:
            self.topBF(value, row)
        # bottom buffer fault condition
        elif (row >= model.start + table_rows):
            self.bottomBF(value, row)
        # We are at top of the dataset
        elif value == self.tricky_vscrollbar.minimum():
            self.vscrollbar.triggerAction(
                QtWidgets.QAbstractSlider.SliderToMinimum)
        # We are at bottom of the dataset
        elif value == self.tricky_vscrollbar.maximum():
            self.vscrollbar.triggerAction(
                QtWidgets.QAbstractSlider.SliderToMaximum)
        # we are somewhere in the middle of the dataset
        else:
            self.scrollTo(
                model.index(row - model.start, 0),
                _aiv.PositionAtTop)

    def topBF(self, value, row):
        """Going out of buffer when browsing upwards.

        Buffer fault condition: row < model.start

        :Parameters:

            - `value`: the current value of the tricky scrollbar
            - `row`: the estimated dataset row mapped to that value
        """

        table_rows = self.tmodel.numrows
        if value == self.tricky_vscrollbar.minimum():
            start = 0
            position = 0
            hint = _aiv.PositionAtTop
            self.vscrollbar.triggerAction(
                QtWidgets.QAbstractSlider.SliderToMinimum)
        else:
            start = row - table_rows
            position = table_rows - 1
            hint = _aiv.PositionAtBottom

        self.tmodel.loadData(start, table_rows)
        self.updateView()
        self.scrollTo(self.tmodel.index(position, 0), hint)

    def bottomBF(self, value, row):
        """Going out of buffer when browsing downwards.

        Buffer fault condition: row > self.tmodel.start + table_rows - 1

        :Parameters:

            - `value`: the current value of the tricky scrollbar
            - `row`: the estimated dataset row mapped to that value
        """

        table_rows = self.tmodel.numrows
        if value == self.tricky_vscrollbar.maximum():
            row = self.leaf_numrows - 1
            start = self.leaf_numrows - table_rows
            position = table_rows - 1
            hint = _aiv.PositionAtBottom
            self.vscrollbar.triggerAction(
                QtWidgets.QAbstractSlider.SliderToMinimum)
        else:
            start = row
            position = 0
            hint = _aiv.PositionAtTop

        self.tmodel.loadData(start, table_rows)
        self.updateView()
        self.scrollTo(self.tmodel.index(position, 0), hint)

    def wheelEvent(self, event):
        """Specialized handler for the wheel events received by the *viewport*.

        :Parameter event: the QWheelEvent being processed
        """

        if self.leaf_numrows > self.tmodel.numrows:
            height = self.vheader.sectionSize(0)
            # The distance the wheel is rotated in eights of a degree.
            # For example: 120/8 = 15 so if delta is 120 then the wheel
            # has been rotated by 15 degrees. It *seems* that every eight of
            # degree corresponds to a distance of 1 pixel.
            delta = event.angleDelta().y()
            self.wheel_step = round(abs(delta) / height) - 1
            if delta < 0:
                self.wheelDown(event)
            else:
                self.wheelUp(event)
            self.syncView()
            # Filter the event so it will not be passed to the parent widget
            event.accept()
        else:
            QtWidgets.QTableView.wheelEvent(self, event)

    def wheelDown(self, event):
        """Setup data for wheeling with the mouse towards the last section.
        """

        model, vh, table_rows, buffer_start, last_vp_row, page_step = \
            self.mouseNavInfo('d')
        vp_rows = last_vp_row - vh.visualIndexAt(0)
        # If we are at the last page of the buffer but not at the last
        # row of the dataset we still can go downwards so we have to
        # read the next contiguous buffer
        if (last_vp_row + self.wheel_step + 1 > table_rows) and \
                (buffer_start + table_rows < self.leaf_numrows):
            # Buffer fault. The new buffer and the old one overlap to ensure
            # that no jumps occur.
            new_start = \
                buffer_start + last_vp_row + self.wheel_step - page_step
            model.loadData(new_start, table_rows)
            self.updateView()
            self.scrollTo(model.index(new_start - model.start, 0),
                          _aiv.PositionAtTop)
        else:
            QtCore.QCoreApplication.sendEvent(self.vscrollbar, event)

    def wheelUp(self, event):
        """Setup data for wheeling with the mouse towards the first section.
        """

        model, vh, table_rows, buffer_start, first_vp_row, page_step = \
            self.mouseNavInfo('u')
        vp_rows = vh.visualIndexAt(self.viewport().height()) - first_vp_row
        # If we are at the first page of the buffer but not at the first
        # row of the dataset we still can go upwards so we have to
        # read the previous contiguous buffer
        if (first_vp_row < page_step + 1) and (buffer_start > 0):
            # Buffer fault. The new buffer and the old one overlap to ensure
            # that no jumps occur.
            new_start = buffer_start + first_vp_row + page_step - \
                self.wheel_step - table_rows + 1
            model.loadData(new_start, table_rows)
            self.updateView()
            self.scrollTo(
                model.index(
                    new_start + table_rows - model.start - 1, 0),
                _aiv.PositionAtBottom)
        else:
            QtCore.QCoreApplication.sendEvent(self.vscrollbar, event)

    def keyPressEvent(self, event):
        """Handle basic cursor movement for key events.

        :Parameter event: the key event being processed
        """

        if self.tmodel.numrows < self.leaf_numrows:
            key = event.key()
            if key == QtCore.Qt.Key_Home:
                event.accept()
                self.homeKeyPressEvent()
            elif key == QtCore.Qt.Key_End:
                event.accept()
                self.endKeyPressEvent()
            elif key == QtCore.Qt.Key_Up:
                event.accept()
                self.upKeyPressEvent(event)
            elif key == QtCore.Qt.Key_Down:
                event.accept()
                self.downKeyPressEvent(event)
            elif key == QtCore.Qt.Key_PageUp:
                event.accept()
                self.pageUpKeyPressEvent(event)
            elif key == QtCore.Qt.Key_PageDown:
                event.accept()
                self.pageDownKeyPressEvent(event)
            else:
                QtWidgets.QTableView.keyPressEvent(self, event)
        else:
            QtWidgets.QTableView.keyPressEvent(self, event)

    def homeKeyPressEvent(self):
        """Specialised handler for the `Home` key press event.

        See enum QAbstractitemView.CursorAction for reference.
        """

        model = self.tmodel
        table_rows = model.numrows
        index = model.index(0, 0)
        # Update buffer if needed
        if model.start > 0:
            model.loadData(0, table_rows)
            self.updateView()
        self.setCurrentIndex(index)
        self.scrollToTop()

        # Eventually synchronize the position of the visible scrollbar
        # the displayed data
        self.tricky_vscrollbar.setValue(0)

    def endKeyPressEvent(self):
        """Specialised handler for the `End` key press event.

        See enum QAbstractitemView.CursorAction for reference.
        """

        model = self.tmodel
        table_rows = model.numrows
        index = model.index(table_rows - 1, model.numcols - 1)
        # Update buffer if needed
        last_row = model.start + table_rows
        if last_row < self.leaf_numrows:
            self.tmodel.loadData(self.leaf_numrows - table_rows,
                                 table_rows)
            self.updateView()
        self.setCurrentIndex(index)
        self.scrollToBottom()

        # Eventually synchronize the position of the visible scrollbar
        # the displayed data
        self.tricky_vscrollbar.setValue(self.max_value)

    def keyboardNavInfo(self):
        """Gives information about model, and current cell.

        This is a helper method used by methods that browse the data via
        keyboard.

        :Parameter direction: the data browsing direction (upwards/downwards)
        """

        model = self.tmodel
        # Load the buffer where the valid current cell lives
        self.validCurrentCellBuffer()

        # About the table
        table_rows = model.numrows
        buffer_start = model.start

        page_step = self.vscrollbar.pageStep()

        # About the current cell
        current_index = self.currentIndex()
        buffer_row = current_index.row()
        buffer_column = current_index.column()
        dataset_row = buffer_start + buffer_row

        return (model, table_rows, buffer_start, page_step, current_index,
                buffer_row, buffer_column, dataset_row)

    def upKeyPressEvent(self, event):
        """Specialised handler for the cursor up key press event.

        :Parameter event: the key event being processed
        """

        (model, table_rows, buffer_start, page_step, current_index,
            buffer_row, buffer_column, dataset_row) = self.keyboardNavInfo()

        # If we are at the first row of the buffer but not at the first
        # row of the dataset we still can go upwards so we have to read
        # the previous contiguous buffer
        if (buffer_row == 0) and (buffer_start > 0):
            model.loadData(dataset_row - table_rows + page_step, table_rows)
            self.updateView()
            # The position of the new current row
            row = dataset_row - model.start - 1
            if row < 0:
                row = 0
            index = model.index(row, buffer_column)
            self.setCurrentIndex(index)
            self.scrollTo(index,
                          _aiv.PositionAtTop)
        else:
            QtWidgets.QTableView.keyPressEvent(self, event)

        # Eventually synchronize the position of the visible scrollbar
        # with the displayed data using the first visible cell as
        # reference
        self.syncView()

    def pageUpKeyPressEvent(self, event):
        """Specialised handler for the `PageUp` key press event.

        :Parameter event: the key event being processed
        """

        (model, table_rows, buffer_start, page_step, current_index,
            buffer_row, buffer_column, dataset_row) = self.keyboardNavInfo()

        # If we are at the first page of the buffer but not at the first
        # page of the dataset we still can go upwards so we have to read
        # the previous contiguous buffer
        if (buffer_row - page_step < 0) and (buffer_start > 0):
            model.loadData(dataset_row - table_rows, table_rows)
            self.updateView()
            # The position of the new current row
            row = dataset_row - model.start - page_step - 1
            if row < 0:
                row = 0
            index = model.index(row, buffer_column)
            self.setCurrentIndex(index)
            self.scrollTo(index,
                          _aiv.PositionAtTop)
        else:
            QtWidgets.QTableView.keyPressEvent(self, event)

        # Eventually synchronize the position of the visible scrollbar
        # with the displayed data using the first visible cell as
        # reference
        self.syncView()

    def downKeyPressEvent(self, event):
        """Specialised handler for the cursor down key press event.

        :Parameter event: the key event being processed
        """

        (model, table_rows, buffer_start, page_step, current_index,
            buffer_row, buffer_column, dataset_row) = self.keyboardNavInfo()

        # If we are at the last row of the buffer but not at the last
        # row of the dataset we still can go downwards so we have to
        # read the next contiguous buffer
        if (buffer_row == table_rows - 1) and \
                (buffer_start + table_rows < self.leaf_numrows):
            model.loadData(dataset_row - page_step + 1, table_rows)
            self.updateView()
            # The position of the new current row
            row = dataset_row - model.start + 1
            if row > table_rows - 1:
                row = table_rows - 1
            index = model.index(row, buffer_column)
            self.setCurrentIndex(index)
            self.scrollTo(index,
                          _aiv.PositionAtBottom)
        else:
            QtWidgets.QTableView.keyPressEvent(self, event)

        # Eventually synchronize the position of the visible scrollbar
        # with the displayed data using the first visible cell as
        # reference
        self.syncView()

    def pageDownKeyPressEvent(self, event):
        """Specialised handler for the `PageDown` key press event.

        :Parameter event: the key event being processed
        """

        (model, table_rows, buffer_start, page_step, current_index,
            buffer_row, buffer_column, dataset_row) = self.keyboardNavInfo()

        # If we are at the last page of the buffer but not at the last
        # row of the dataset we still can go downwards so we have to
        # read the next contiguous buffer
        if (buffer_row + page_step > table_rows - 1) and \
                (buffer_start + table_rows < self.leaf_numrows):
            model.loadData(dataset_row + 1, table_rows)
            self.updateView()
            # The position of the new current row
            row = dataset_row - model.start + page_step + 1
            if row > table_rows - 1:
                row = table_rows - 1
            index = model.index(row, buffer_column)
            self.setCurrentIndex(index)
            self.scrollTo(index,
                          _aiv.PositionAtBottom)
        else:
            QtWidgets.QTableView.keyPressEvent(self, event)

        # Eventually synchronize the position of the visible scrollbar
        # with the displayed data using the first visible cell as
        # reference
        self.syncView()

    # For large datasets the number of rows of the dataset is greater than
    # the number of rows of the table used for displaying data. It means
    # that if a table cell is activated (made selected by clicking, double
    # cliking or pressing the arrow keys. We call this cell *valid current
    # cell*) we must take care for keeping it updated. For instance, if
    # the table contains the first 10000 rows of the dataset and the user
    # clicks the cell (52, 1) it becomes the valid current cell. If now the
    # user draggs the slider and goes down let say 30000 rows then the row
    # 52 of the table will still be the current one but its section number
    # will be 30052 which is wrong (the current cell can only be changed by
    # activating other cell, not by dragging the scrollbar). We call this
    # cell *fake current cell*. This would be a bug.

    def currentChanged(self, current, previous):
        """Track the dataset current cell.

        This SLOT is automatically called when the current cell changes.

        :Parameters:

        - `current`: the new current index
        - `previous`: the previous current index
        """

        QtWidgets.QTableView.currentChanged(self, current, previous)
        if self.tmodel.numrows < self.leaf_numrows:
            self.valid_current_buffer = self.tmodel.start

    # This method has been renamed from loadDatasetCurrentCell to
    # validCurrentCellBuffer. The method has been debugged too
    def validCurrentCellBuffer(self):
        """Load the buffer in which the valid current cell lives.
        """

        table_rows = self.tmodel.numrows
        valid_current = self.currentIndex().row() + self.valid_current_buffer
        if not (self.tmodel.start <=
                valid_current <=
                self.tmodel.start + table_rows - 1):
            self.tmodel.loadData(self.valid_current_buffer, table_rows)
            self.updateView()

    def selectionChanged(self, selected, deselected):
        """Track the dataset selected cells.

        This method is automatically called when the selected range changes.

        :Parameters:

        - `selected`: the new selection
        - `deselected`: the previous selection (maybe empty)
        """

        model = self.tmodel
        if model.numrows < self.leaf_numrows:
            # Get the selected indexes from the QItemSelection object
            selection = selected.indexes()
            if len(selection):
                model.selected_cell = {
                    'index': selection[0],
                    'buffer_start': model.start,
                }
        else:
            QtWidgets.QTableView.selectionChanged(self, selected, deselected)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
This module displays in a dialog the node information collected by the
:mod:`vitables.nodeprops.nodeinfo` module.

Users' attributes can be edited if the database has been opened in read-write
mode. Otherwise all shown information is read-only.
"""

import logging
import os.path
from vitables.nodeprops import attreditor
import vitables.utils

from qtpy import QtCore
from qtpy import QtGui
from qtpy import QtWidgets
from qtpy.uic import loadUiType
import tables


__docformat__ = 'restructuredtext'

translate = QtWidgets.QApplication.translate
# This method of the PyQt5.uic module allows for dynamically loading user
# interfaces created by QtDesigner. See the PyQt5 Reference Guide for more
# info.
Ui_AttrPropDialog = \
    loadUiType(os.path.join(os.path.dirname(__file__), 'attr_prop_dlg.ui'))[0]


log = logging.getLogger(__name__)


class AttrPropDlg(QtWidgets.QDialog, Ui_AttrPropDialog):
    """
    Node properties dialog.

    By loading UI files at runtime we can:

        - create user interfaces at runtime (without using pyuic)
        - use multiple inheritance, MyParentClass(BaseClass, FormClass)

    This class displays a tabbed dialog that shows some properties of
    the selected node. First tab, General, shows general properties like
    name, path, type etc. The second and third tabs show the system and
    user attributes in a tabular way.

    Beware that data types shown in the General page are `PyTables` data
    types so we can deal with `bytes`, `enum`, `time64` and `pseudoatoms`
    (not all of them are supported by ``numpy``).
    However data types shown in the System and User attributes pages are
    ``numpy`` data types because `PyTables` attributes are stored as ``numpy``
    arrays.

    :Parameter info: a :meth:`vitables.nodeprops.nodeinfo.NodeInfo` instance
      describing a given node
    """

    def __init__(self, info):
        """Setup the Properties dialog."""

        vtapp = vitables.utils.getVTApp()
        super(AttrPropDlg, self).__init__(vtapp.gui)
        self.setupUi(self)

        # Customise the dialog's pages
        self.fillSysAttrsPage(info)
        self.fillUserAttrsPage(info)
        self.resize(self.size().width(), self.minimumHeight())

        # Variables used for checking the table of user attributes
        self.mode = info.mode
        self.asi = info.asi

    def fillSysAttrsPage(self, info):
        """Fill the page of system attributes.

        :Parameter info: a :meth:`vitables.nodeprops.nodeinfo.NodeInfo` instance
          describing a given node
        """

        # Number of attributes label
        self.sattrLE.setText(str(len(info.system_attrs)))

        # Table of system attributes
        self.sysTable.horizontalHeader().setSectionResizeMode(
            QtWidgets.QHeaderView.Stretch)
        self.sysattr_model = QtGui.QStandardItemModel()
        self.sysattr_model.setHorizontalHeaderLabels([
            translate('AttrPropDlg', 'Name',
                      'First column header of the table'),
            translate('AttrPropDlg', 'Value',
                      'Second column header of the table'),
            translate('AttrPropDlg', 'Datatype',
                      'Third column header of the table')])
        self.sysTable.setModel(self.sysattr_model)

        # Fill the table
        bg_brush = self.sysTable.palette().brush(QtGui.QPalette.Window)
        base_brush = self.sysTable.palette().brush(QtGui.QPalette.Base)
        for name, value in info.system_attrs.items():
            name_item = QtGui.QStandardItem(name)
            name_item.setEditable(False)
            name_item.setBackground(bg_brush)
            # Find out the attribute datatype.
            # Since PyTables1.1 scalar attributes are stored as numarray arrays
            # Since PyTables2.0 scalar attributes are stored as numpy arrays
            # It includes the TITLE attribute
            # For instance, assume the ASI of a given node is asi. Then
            # if I do asi.test = 3 ->
            # type(asi.test) returns numpy.int32
            # isinstance(asi.test, int) returns True
            # asi.test.shape returns ()
            # asi.test2 = "hello" ->
            # type(asi.test2) returns numpy.string_
            # isinstance(asi.test2, str) returns True
            # asi.test2.shape returns ()
            # Beware that objects whose shape is () are not warranted
            # to be Python objects, for instance
            # x = numpy.array(3) ->
            # x.shape returns ()
            # type(x) returns numpy.ndarray
            # isinstance(x, int) returns False
            if isinstance(value, tables.Filters):
                dtype_name = 'tables.filters.Filters'
            elif hasattr(value, 'shape'):
                dtype_name = value.dtype.name
            else:
                # Attributes can be scalar Python objects (PyTables <1.1)
                # or non scalar Python objects, e.g. sequences
                dtype_name = str(type(value))
            dtype_item = QtGui.QStandardItem(dtype_name)
            dtype_item.setEditable(False)
            dtype_item.setBackground(bg_brush)
            value_item = QtGui.QStandardItem(str(value))
            value_item.setEditable(False)
            value_item.setBackground(bg_brush)
            # When the database is in read-only mode the TITLE attribute
            # cannot be edited
            if (name == 'TITLE') and (info.mode != 'read-only'):
                # The position of the TITLE value in the table
                self.title_row = self.sysattr_model.rowCount()
                self.title_before = value_item.text()
                value_item.setEditable(True)
                value_item.setBackground(base_brush)
            self.sysattr_model.appendRow([name_item, value_item, dtype_item])

    def fillUserAttrsPage(self, info):
        """Fill the page of user attributes.

        :Parameter info: a :meth:`vitables.nodeprops.nodeinfo.NodeInfo` instance
          describing a given node
        """

        self.user_attrs_before = []

        # Number of attributes label
        self.uattrLE.setText(str(len(info.user_attrs)))

        # Table of user attributes
        self.userTable.horizontalHeader().\
            setSectionResizeMode(QtWidgets.QHeaderView.Stretch)
        self.userattr_model = QtGui.QStandardItemModel()
        self.userattr_model.setHorizontalHeaderLabels([
            translate('AttrPropDlg', 'Name',
                      'First column header of the table'),
            translate('AttrPropDlg', 'Value',
                      'Second column header of the table'),
            translate('AttrPropDlg', 'Datatype',
                      'Third column header of the table')])
        self.userTable.setModel(self.userattr_model)

        # Fill the table
        # The Data Type cell is a combobox with static content
        dtypes_list = ['int8', 'int16', 'int32', 'int64', 'uint8', 'uint16',
                       'uint32', 'uint64', 'float32', 'float64', 'complex64',
                       'complex128', 'bool', 'bytes', 'string', 'python']

        # Brushes used for read-only and writable cells
        bg_brush = self.userTable.palette().brush(QtGui.QPalette.Window)
        base_brush = self.userTable.palette().brush(QtGui.QPalette.Base)
        for name, value in info.user_attrs.items():
            name_item = QtGui.QStandardItem(name)
            value_item = QtGui.QStandardItem(str(value))
            dtype_item = QtGui.QStandardItem()
            dtypes_combo = QtWidgets.QComboBox()
            dtypes_combo.addItems(dtypes_list)
            dtypes_combo.setEditable(False)
            # In PyTables >=1.1 scalar attributes are stored as numarray arrays
            # In PyTables >= 2.0 scalar attributes are stored as numpy arrays
            if hasattr(value, 'shape'):
                dtype_name = value.dtype.name
                if dtype_name.startswith('bytes'):
                    dtype_name = 'bytes'
                elif dtype_name.startswith('str'):
                    dtype_name = 'string'
            else:
                # Attributes can be scalar Python objects (PyTables <1.1)
                # or non scalar Python objects, e.g. sequences
                dtype_name = 'python'
            self.userattr_model.appendRow([name_item, value_item, dtype_item])
            dtypes_combo.setCurrentIndex(dtypes_combo.findText(dtype_name))
            self.userTable.setIndexWidget(dtype_item.index(), dtypes_combo)

            # Complex attributes and ND_array attributes need some visual
            # adjustments
            if dtype_name.startswith('complex'):
                # Remove parenthesis from the str representation of
                # complex numbers.
                if (str(value)[0], str(value)[-1]) == ('(', ')'):
                    value_item.setText(str(value)[1:-1])
            # ViTables doesn't support editing ND-array attributes so
            # they are displayed in non editable cells
            if (hasattr(value, 'shape') and value.shape != ())or\
                    (info.mode == 'read-only'):
                editable = False
                brush = bg_brush
            else:
                editable = True
                brush = base_brush
            name_item.setEditable(editable)
            name_item.setBackground(brush)
            value_item.setEditable(editable)
            value_item.setBackground(brush)
            self.user_attrs_before.append((name_item.text(), value_item.text(),
                                           dtypes_combo.currentText()))
        self.user_attrs_before = sorted(self.user_attrs_before[:])

        # The group of buttons Add, Delete, What's This
        self.page_buttons = QtWidgets.QButtonGroup(self.userattrs_page)
        self.page_buttons.addButton(self.addButton, 0)
        self.page_buttons.addButton(self.delButton, 1)
        self.page_buttons.addButton(self.helpButton, 2)

        # If the database is in read-only mode user attributes cannot be edited
        if info.mode == 'read-only':
            for uid in (0, 1):
                self.page_buttons.button(uid).setEnabled(False)

        self.helpButton.clicked.connect(
            QtWidgets.QWhatsThis.enterWhatsThisMode)

    @QtCore.Slot("QModelIndex", name="on_sysTable_clicked")
    @QtCore.Slot("QModelIndex", name="on_userTable_clicked")
    def displaySelectedCell(self, index):
        """Show the content of the clicked cell in the line edit at bottom.

        :Parameter index: the model index of the clicked cell
        """

        page = self.tabw.currentIndex()
        if page == 1:
            model_item = self.sysattr_model.itemFromIndex(index)
            self.systemAttrCellLE.clear()
            self.systemAttrCellLE.setText(model_item.text())
        elif page == 2:
            model_item = self.userattr_model.itemFromIndex(index)
            self.userAttrCellLE.clear()
            self.userAttrCellLE.setText(model_item.text())

    @QtCore.Slot(name="on_addButton_clicked")
    def addAttribute(self):
        """Add a new attribute to the user's attributes table."""

        # The Data Type cell is a combobox with static content
        dtypes_list = ['int8', 'int16', 'int32', 'int64', 'uint8', 'uint16',
                       'uint32', 'uint64', 'float32', 'float64', 'complex64',
                       'complex128', 'bool', 'bytes', 'string', 'python']

        name_item = QtGui.QStandardItem()
        value_item = QtGui.QStandardItem()
        dtype_item = QtGui.QStandardItem()
        dtypes_combo = QtWidgets.QComboBox()
        dtypes_combo.addItems(dtypes_list)
        dtypes_combo.setEditable(False)
        self.userattr_model.appendRow([name_item, value_item, dtype_item])
        self.userTable.setIndexWidget(dtype_item.index(), dtypes_combo)

        # Start editing the proper cell. If not, clicking Add+Delete
        # would result in the deletion of an attribute different to that
        # just added. It is also more handy as it allows to start editing
        # without double clicking the cell.
        self.userTable.edit(name_item.index())

    @QtCore.Slot(name="on_delButton_clicked")
    def delAttribute(self):
        """
        Remove an attribute from the user's attributes table.

        This slot is connected to the clicked signal of the `Delete` button.
        An attribute is marked for deletion by giving focus to any cell
        of the row describing it (i.e. clicking a cell or selecting its
        contents).
        """

        # If there is not a selected attribute then return
        current_index = self.userTable.currentIndex()
        if not current_index.isValid():
            log.error(
                translate('AttrPropDlg',
                          'Please, select the attribute to be deleted.',
                          'A usage text'))
            return

        # Get the name of the attribute being deleted
        current_row = current_index.row()
        current_column = current_index.column()
        if current_column != 0:
            current_index = current_index.sibling(current_row, 0)
        name = self.userattr_model.itemFromIndex(current_index).text()

        # Delete the marked attribute
        title = translate('AttrPropDlg', 'User attribute deletion',
                          'Caption of the attr deletion dialog')
        text = translate('AttrPropDlg',
                         "\n\nYou are about to delete the attribute:\n{0}\n\n",
                         'Ask for confirmation').format(name)
        itext = ''
        dtext = ''
        buttons = {
            'Delete':
                (translate('AttrPropDlg', 'Delete', 'Button text'),
                 QtWidgets.QMessageBox.YesRole),
            'Cancel':
                (translate('AttrPropDlg', 'Cancel', 'Button text'),
                 QtWidgets.QMessageBox.NoRole),
        }

        # Ask for confirmation
        answer = vitables.utils.questionBox(title, text, itext, dtext, buttons)
        if answer == 'Delete':
            # Updates the user attributes table
            self.userattr_model.removeRow(current_row)

    def asiChanged(self):
        """Find out if the user has edited some attribute.

        In order to decide if attributes have been edited we compare
        the attribute tables contents at dialog creation time and at
        dialog closing time.
        """

        # Get the value of the TITLE attribute. Checking is required
        # because the attribute is mandatory in PyTables but not in
        # generic HDF5 files
        self.title_after = None
        if hasattr(self, 'title_row'):
            self.title_after = \
                self.sysattr_model.item(self.title_row, 1).text()
            if self.title_before != self.title_after:
                return True

        # Check user attributes
        self.user_attrs_after = []
        for index in range(0, self.userTable.model().rowCount()):
            name_after = self.userattr_model.item(index, 0).text()
            value_after = self.userattr_model.item(index, 1).text()
            dtype_item = self.userattr_model.item(index, 2)
            dtype_combo = self.userTable.indexWidget(dtype_item.index())
            dtype_after = dtype_combo.currentText()
            self.user_attrs_after.append((name_after, value_after,
                                          dtype_after))
        if sorted(self.user_attrs_before) != sorted(self.user_attrs_after):
            return True

        return False

    @QtCore.Slot(name="on_buttonsBox_accepted")
    def accept(self):
        """
        Overwritten slot for accepted dialogs.

        This slot is always the last called method whenever the Apply/Ok
        buttons are pressed.
        """

        # If the file is in read-only mode or the Attribute Set Instance
        # remains unchanged no attribute needs to be updated
        if (self.mode == 'read-only') or (not self.asiChanged()):
            QtWidgets.QDialog.accept(self)
            return  # This is mandatory!

        # Check the editable attributes
        aeditor = attreditor.AttrEditor(self.asi, self.title_after,
                                        self.userTable)
        attrs_are_ok, error = aeditor.checkAttributes()
        # If the attributes pass correctness checks then update the
        # attributes and close the dialog
        if attrs_are_ok is True:
            aeditor.setAttributes()
            del aeditor
            QtWidgets.QDialog.accept(self)
            return
        # If not then keep the dialog opened
        else:
            del aeditor
            self.tabw.setCurrentIndex(2)
            self.userAttrCellLE.clear()
            self.userAttrCellLE.setText(error)



"""Configuration file for our tests.

Here we create fixtures that are not directly required by test functions.
"""

import sys

import pytest

import tables

from qtpy import QtWidgets

import vitables.vtapp
from vitables.preferences import vtconfig


class Launcher(object):
    def __init__(self):
        self.app = QtWidgets.QApplication(sys.argv)
        self.app.setOrganizationDomain('vitables.org')
        self.app.setOrganizationName('ViTables')
        self.app.setApplicationName('ViTables')
        self.app.setApplicationVersion(vtconfig.getVersion())
        self.vtapp_object = vitables.vtapp.VTApp(keep_splash=False)
        self.gui = self.vtapp_object.gui


@pytest.fixture(scope='module')
def launcher():
    return Launcher()

@pytest.fixture(scope='module')
def h5file():
    return tables.open_file('testfile.h5', 'r')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""Plugin that provides nice string formatting for time fields.

It supports not only native `PyTables` time datatypes but also the time
series contained in `PyTables` tables generated via ``scikits.timeseries`` and
``Pandas``.
"""

import time
import os
try:
    import configparser
except ImportError:
    import ConfigParser as configparser

import tables

try:
    import scikits.timeseries as ts
except ImportError:
    ts = None

try:
    import pandas as pd
except ImportError:
    pd = None

from qtpy import QtCore, QtGui
from qtpy.QtCore import Qt
from qtpy import QtWidgets

import vitables.utils
from vitables.plugins.timeseries.aboutpage import AboutPage

_axis_font = QtGui.QFont()
_axis_font.setBold(True)

__docformat__ = 'restructuredtext'
__version__ = '2.1'
plugin_name = 'Time series formatter'
comment = 'Display time series in a human friendly format'

translate = QtWidgets.QApplication.translate


def findTS(leaf, node_kind):
    """Find out if the inspected leaf contains a time field.

    **Existing time fields that cannot be formatted are skipped**:

        - time series created via ``pandas`` module are ignored if that
          module is not available
        - time series created via ``scikits.timeseries`` module are
          ignored if that module is not available
        - time fields that are displayed in a multidimensional cell
          are ignored

    The last restriction includes the following cases:

        - time fields that are part of a nested field
        - time fields in `VLArrays`
        - time fields in arrays with more than 2 dimensions
        - time fields in arrays with atom shape other than ()

    :Parameters:
        - `leaf`: the tables.Leaf instance being inspected.
        - `node_kind`: a LeafNode attribute that indicates the kind of dataset

    :Return ts_kind: a flag indicating the kind of time series found
    """

    time_types = ['time32', 'time64']
    if isinstance(leaf, tables.Table):
        asi = leaf._v_attrs
        coltypes = leaf.coltypes
        pgroup = leaf._g_getparent()

        # Check for Pandas timeseries
        pandas_attr = getattr(pgroup._v_attrs, 'pandas_type', None)
        if pd and pandas_attr in ['frame', 'frame_table']:
            userattrs_names = asi._v_attrnamesuser
            dtype_attrs = [n for n in userattrs_names if n.endswith('_dtype')]
            dtype_attrs.append('index_kind')
            for n in dtype_attrs:
                if getattr(asi, n).startswith('datetime'):
                    return 'pandas_ts'

        # Check for scikits.timeseries timeseries
        if ts and hasattr(asi, 'CLASS') and (asi.CLASS == 'TimeSeriesTable'):
            return 'scikits_ts'

        # Check for PyTables timeseries
        for name in leaf.colnames:
            if (name in coltypes) and (coltypes[name] in time_types):
                return 'pytables_ts'
    elif (leaf.atom.type in time_types) and \
            (len(leaf.shape) < 3) and \
            (leaf.atom.shape == ()) and \
            (node_kind != 'vlarray'):
        return 'pytables_ts'

    return None


def tsPositions(ts_kind, leaf):
    """Return the position of the time field going to be formatted.

    The following cases can occur:

    - scikits_ts: leaf is a `TimeSeriesTable` instance. It contains just one
      time field, in a column labeled as `_dates`.
    - pandas_ts: leaf is a regular `tables.Table` instance with a column
      named `index`.
    - pytables_ts and leaf is a regular `tables.Table` instance. Every column
      can contain a time data type so we must inspect every column
    - pytables_ts and leaf is a `tables.Array` instance. As it is a homogeneous
      data container if a column contains a time data type then every column
      contains a time data type. Specific positions are not required.
    """

    positions = []
    if (ts_kind == 'scikits_ts'):
        positions.append(leaf.coldescrs['_dates']._v_pos)
    elif (ts_kind == 'pandas_ts'):
        hstore = pd.HDFStore(leaf._v_file.filename)
        pgroup = leaf._g_getparent()
        df = hstore[pgroup._v_name]

        # Inspect dataframe index
        df_index = df.index
        if isinstance(df_index, pd.MultiIndex):
            nlevels = df_index.nlevels
            for i, idx in enumerate(df_index.names):
                dtime = 'datetime'
                if df_index.get_level_values(i).dtype.name.startswith(dtime):
                    positions.append(i)
        elif df_index.dtype.name.startswith('datetime'):
            nlevels = 1
            positions.append(0)

        # Inspect dataframe data
        for i, dt in enumerate(df.dtypes):
            if dt.name.startswith('datetime'):
                positions.append(nlevels + i)
    elif ts_kind == 'pytables_ts':
        if isinstance(leaf, tables.Table):
            for name in leaf.colnames:
                if leaf.coltypes[name] in ['time32', 'time64']:
                    positions.append(leaf.coldescrs[name]._v_pos)
        else:
            positions = [-1]
    return positions


def tsFrequency(ts_kind, leaf):
    """Return the frequency (if any) of the time series.

    Only time series created via ``scikits.timeseries`` module have
    this attribute.
    """

    ts_freq = None
    if ts_kind == 'scikits_ts':
        # The frequency of the time serie. Default is 6000 (daily)
        special_attrs = getattr(leaf._v_attrs, 'special_attrs',
                                {'freq': 6000})
        ts_freq = special_attrs['freq']
    return ts_freq


def datetimeFormat():
    """The format string to be used when rendering the time series.
    """

    config = configparser.ConfigParser(interpolation=None)
    ini_filename = os.path.join(os.path.dirname(__file__), 'time_format.ini')
    def_dtformat = '%c'
    try:
        config.read_file(open(ini_filename))
        datetime_format = config['Timeseries']['strftime']
    except (IOError, configparser.ParsingError):
        datetime_format = def_dtformat

    return datetime_format


class TSFormatter(object):
    """Human friendly formatting of time series in a dataset.

    An inspector class intended for finding out if a `tables.Leaf` instance
    contains a time series suitable to be formatted in a user friendly way.
    """

    UID = __name__
    NAME = plugin_name
    COMMENT = comment

    def __init__(self):
        """Class constructor.

        Dynamically finds new instances of
        :meth:`vitables.vttables.leaf_model.LeafModel` and customises them if
        they contain time fields.
        """

        self.vtapp = vitables.utils.getVTApp()
        self.vtapp.leaf_model_created.connect(self.customiseModel)

    def customiseModel(self, datasheet):
        """Inspect a leaf model and customise it if a time series is found.

        :Parameter subwindow: the :meth:`vitables.vttables.datasheet.DataSheet`
          instance being inspected
        """

        # If the node is a soft/external link then dereference it
        leaf = datasheet.dbt_leaf.node
        if isinstance(leaf, tables.link.Link):
            leaf = leaf.__call__()

        # Look for formattable time fields in the dataset
        node_kind = datasheet.dbt_leaf.node_kind
        ts_kind = findTS(leaf, node_kind)
        if ts_kind is None:
            return

        # Get the positions of the time fields
        time_cols = tsPositions(ts_kind, leaf)
        if time_cols == []:
            return

        # Customise the leaf model
        model = datasheet.widget().model()
        ts_info = {
            'ts_kind': ts_kind,
            'ts_cols': time_cols,
            'ts_freq': tsFrequency(ts_kind, leaf),
            'ts_format': datetimeFormat(),
        }
        if isinstance(leaf, tables.Table):
            if isinstance(model, vitables.vttables.df_model.DataFrameModel):
                leaf_kind = 'dataframe'
            else:
                leaf_kind = 'table'
        else:
            leaf_kind = 'array'
        model_info = {
            'leaf_kind': leaf_kind,
            'model': model,
            'numrows': model.rowCount(),
        }
        if leaf_kind in ['table', 'array']:
            model_info['formatContent'] = model.formatContent

        # Add required attributes to model
        for k in ts_info:
            setattr(model, k, ts_info[k])

        # Add/customise required methods to model
        ts_model = TSLeafModel(model_info, ts_info)
        model.tsFormatter = ts_model.tsFormatter
        model.data = ts_model.data

    def helpAbout(self, parent):
        """Full description of the plugin.

        This is a convenience method which works as expected by
        :meth:preferences.preferences.Preferences.aboutPluginPage i.e.
        build a page which contains the full description of the plugin
        and, optionally, allows for its configuration.

        :Parameter about_page: the container widget for the page
        """

        # Plugin full description
        desc = {'version': __version__,
                'module_name': os.path.join(os.path.basename(__file__)),
                'folder': os.path.join(os.path.dirname(__file__)),
                'author': 'Vicent Mas <vmas@vitables.org>',
                'about_text': translate('TimeFormatterPage',
                                        """<qt>
            <p>Plugin that provides nice string formatting for time fields.
            <p>It supports not only native PyTables time datatypes but
            also time series generated (and stored in PyTables tables) via
            Pandas and scikits.timeseries packages.
            </qt>""",
                                        'Text of an About plugin message box')}
        self.about_page = AboutPage(desc, parent)

        # We need to install the event filter because the Preferences dialog
        # eats all Return key presses even if the time format editor widget
        # has the keyboard focus (so connecting the returnPressed signal
        # of this widget to the AboutPage.applyFormat is useless)
        parent.parent().installEventFilter(self.about_page)
        return self.about_page


class TSLeafModel(object):
    """Provides a `data()` method to leaf models that contains time series.

    Formatting a table is more difficult than formatting an array because
    tables content is not homogeneous and columns with time series have to
    be formatted in a different way to the rest of columns.
    """

    def __init__(self, model_info, ts_info, parent=None):
        """The constructor method.

        All required attributes are set in this method.
        """

        # Attributes required by the tsFormatter() method
        self.ts_kind = ts_info['ts_kind']
        self.ts_freq = ts_info['ts_freq']
        self.ts_format = ts_info['ts_format']
        # Attributes required by the data() method
        self.model = model_info['model']
        self.numrows = model_info['numrows']
        self.ts_cols = ts_info['ts_cols']
        if 'formatContent' in model_info:
            self.formatContent = model_info['formatContent']

        self.tsFormatter = self.timeFormatter()

        leaf_kind = model_info['leaf_kind']
        if leaf_kind == 'table':
            self.data = self.table_data
        elif leaf_kind == 'dataframe':
            self.data = self.df_data
        else:
            self.data = self.array_data

    def table_data(self, index, role=QtCore.Qt.DisplayRole):
        """Returns the data stored under the given role for the item
        referred to by the index.

        This is an overwritten method.

        :Parameters:

        - `index`: the index of a data item
        - `role`: the role being returned
        """
        row, col = index.row(), index.column()

        if not index.isValid() or not (0 <= row < self.numrows):
            return None

        if role == QtCore.Qt.DisplayRole:
            cell = self.model.cell(row, col)
            if index.column() in self.ts_cols:
                return self.tsFormatter(cell)
            return self.formatContent(cell)

        if role == QtCore.Qt.TextAlignmentRole:
            return QtCore.Qt.AlignLeft | QtCore.Qt.AlignTop

        return None

    def df_data(self, index, role=QtCore.Qt.DisplayRole):
        """Returns the data stored under the given role for the item
        referred to by the index.

        This is an overwritten method.

        :Parameters:

        - `index`: the index of a data item
        - `role`: the role being returned
        """
        row, col = index.row(), index.column()
        n_columns, n_index = self.model._nheaders
        df = self.model._chunk

        if not index.isValid() or not (0 <= row < (self.numrows + n_columns)):
            return None

        is_index = col < n_index
        is_columns = (self.model.start + row) < n_columns

        if is_index and is_columns:
            return None

        if is_index:
            if role == Qt.DisplayRole:
                val = df.index[row - n_columns]
                if n_index > 1:
                    val = val[col]
                if col in self.ts_cols:
                    return self.tsFormatter(val)
                return str(val)
            if role == Qt.FontRole:
                return _axis_font
            if role == Qt.TextAlignmentRole:
                return int(Qt.AlignRight | Qt.AlignVCenter)
            return

        if is_columns:
            if role == Qt.DisplayRole:
                val = df.columns[col - n_index]
                if n_columns > 1:
                    val = val[row]
                return str(val)
            if role == Qt.FontRole:
                return _axis_font
            if role == Qt.TextAlignmentRole:
                return int(Qt.AlignCenter | Qt.AlignBottom)
            return

        if role == Qt.DisplayRole:
            val = self.model._chunk.iat[row - n_columns, col - n_index]
            if col in self.ts_cols:
                return self.tsFormatter(val)
            return str(val)

        # if role == Qt.TextAlignmentRole:
        #     return int(Qt.AlignLeft|Qt.AlignTop)

        return

    def array_data(self, index, role=QtCore.Qt.DisplayRole):
        """Returns the data stored under the given role for the item
        referred to by the index.

        This is an overwritten method.

        :Parameters:

        - `index`: the index of a data item
        - `role`: the role being returned
        """
        row, col = index.row(), index.column()

        if not index.isValid() or not (0 <= row < self.numrows):
            return None

        if role == QtCore.Qt.DisplayRole:
            cell = self.model.cell(row, col)
            return self.tsFormatter(cell)

        if role == QtCore.Qt.TextAlignmentRole:
            return QtCore.Qt.AlignLeft | QtCore.Qt.AlignTop

        return None

    def timeFormatter(self):
        """Return the function to be used for formatting time series.
        """

        time_formatter = None
        ts_kind = self.ts_kind
        if ts_kind == 'pandas_ts':
            time_formatter = self.formatPandasTS
        elif ts_kind == 'scikits_ts':
            time_formatter = self.formatScikitsTS
        elif ts_kind == 'pytables_ts':
            time_formatter = self.formatPyTablesTS
        return time_formatter

    def formatPyTablesTS(self, content):
        """
        Format a given date in a user friendly way.

        The textual representation of the date index is converted to a UTC
        time that can be easily formatted. This method is called when the
        timeseries has not been created using a third party library (i.e;
        Pandas, scikits.timeseries packages).
        """

        try:
            return time.strftime(self.ts_format, time.gmtime(content))
        except ValueError:
            return content

    def formatPandasTS(self, content):
        """Format a given date in a user friendly way.

        The textual representation of the date index is converted to a
        Timestamp instance that can be easily formatted.

        :Parameter content: the content of the table cell being formatted
        """
        try:
            return content.strftime(self.ts_format)
        except ValueError:
            return content

    def formatScikitsTS(self, content):
        """Format a given date in a user friendly way.

        The textual representation of the date index is converted to a Date
        instance that can be easily formatted.

        :Parameter content: the content of the table cell being formatted
        """

        date = ts.Date(self.ts_freq, value=int(content))
        try:
            return date.datetime.strftime(self.ts_format)
        except ValueError:
            return content



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""This module provides a console (named Logger) to ``ViTables``.

Warning and error messages delivered to users by the application are
displayed in this console. Also some messages about the result of the
operations requested by users are shown in the console. This is a read-only
console, its main purpose is to feed users with important information
regarding its working session (so it is called Logger). Users can't enter
commands in the Logger.

    :Parameter parent: the parent widget of the Logger
"""

from qtpy import QtCore
from qtpy import QtGui
from qtpy import QtWidgets

import vitables.utils

__docformat__ = 'restructuredtext'

translate = QtWidgets.QApplication.translate


class Logger(QtWidgets.QTextEdit):
    """
    Console that receives all informational application messages.

    All messages delivered by application to user are displayed in the
    Logger. This messages include the status of user requested
    operations, the result of this operations, and also error messages.
    This is possible because the class reimplement the write() method,
    so it can be used to catch both, ``sys.stdout`` and ``sys.stderr``, just by
    redirecting them to an instance of this class.
    """

    def __init__(self, parent=None):
        """Create the Logger widget and configure it.
        """

        super(Logger, self).__init__(parent)
        self.setAcceptRichText(True)
        self.setReadOnly(1)
        self.setMinimumHeight(50)
        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.setWhatsThis(
            translate(
                'Logger',
                """<qt>
                <h3>The Logger</h3>
                This screen region is a read-only console that logs the result
                of operations requested by the user.
                <p>Also execution errors and exceptions are logged in this
                console.</qt>""",
                'WhatsThis help for the logger'))

        self.setStyleSheet("background-color: #ffffff")

        # The frame especification
        self.frame_style = {'shape': self.frameShape(),
                            'shadow': self.frameShadow(),
                            'lwidth': self.lineWidth()}

        # The context menu
        self.context_menu = self.setupContextMenu()

        # Connect signals to slots
        self.customContextMenuRequested.connect(self.popupCustomContextMenu)

    def write(self, text):
        """
        Catch the messages sent to the standard output and the standard error.

        The caught messages are displayed in the `Logger` using this method.

        The implementation is done via QTextEdit.append() method because it
        adds a new paragraph at the end of the console content so even if
        the user clicks somewhere in the console or there is selected text, the
        new text will not mess the console.

        :Parameter text: the text being written
        """

        current_color = self.textColor()
        if text in ['\n', '\r\n']:
            return
        if text.startswith('\nError: '):
            self.setTextColor(QtGui.QColor('red'))
        elif text.startswith('\nWarning: '):
            self.setTextColor(QtGui.QColor(243, 137, 8))
        self.append(text)
        # Reset the text color
        self.setTextColor(current_color)

    def flush(self):
        """
        Stopgap function to allow use of this class with logging.StreamHandler.
        """
        pass

    def setupContextMenu(self):
        """Create a customised context menu."""

        # Make the menu
        edit_menu = QtWidgets.QMenu(self)
        edit_menu.setObjectName('logger_context_menu')

        # QtGui.Qpalette.Window constant is 10
        edit_menu.setStyleSheet("background-color: {0}".format(10))

        self.copy_action = QtWidgets.QAction(
            translate('Logger', "&Copy", 'Logger menu entry'), self,
            shortcut=QtGui.QKeySequence.Copy, triggered=self.parent().makeCopy,
            statusTip=translate('Logger', 'Copy selected text to clipboard',
                                'Status bar text for the logger context menu ->'
                                ' Copy action'))
        self.copy_action.setObjectName('logger_copy_action')
        edit_menu.addAction(self.copy_action)

        self.clear_action = QtWidgets.QAction(
            translate('Logger', "Cl&ear All", 'Logger menu entry'), self,
            triggered=self.clear,
            statusTip=translate('Logger', 'Empty the Logger',
                                'Status bar text for the logger context menu ->'
                                ' Clear action'))
        self.clear_action.setObjectName('logger_clear_action')
        edit_menu.addAction(self.clear_action)
        edit_menu.addSeparator()

        self.select_action = QtWidgets.QAction(
            translate('Logger', "Select &All", 'Logger menu entry'), self,
            triggered=self.selectAll,
            statusTip=translate('Logger', 'Select the whole Logger contents',
                                'Status bar text for the logger context menu ->'
                                ' Select All'))
        self.select_action.setObjectName('logger_select_action')
        edit_menu.addAction(self.select_action)

        edit_menu.aboutToShow.connect(self.updateContextMenu)
        return edit_menu

    def popupCustomContextMenu(self, pos):
        """
        Popup the context logger menu.

        :Parameter pos: the local position at which the menu will popup
        """
        self.context_menu.popup(self.mapToGlobal(pos))

    def updateContextMenu(self):
        """Update items availability when the menu is about to be shown."""

        # Copy is enabled when there is some selected text
        cursor = self.textCursor()
        self.copy_action.setEnabled(cursor.hasSelection())

        # Clear All and Select All are enabled when there is some text
        has_content = not self.document().isEmpty()
        self.clear_action.setEnabled(has_content)
        self.select_action.setEnabled(has_content)

    def focusInEvent(self, event):
        """Specialised handler for focus events.

        Repaint differently the `Logger` frame when it gets the keyboard focus
        so that users can realize easily about this focus change.

        :Parameter event: the event being processed
        """

        self.setLineWidth(2)
        self.setFrameShape(QtWidgets.QFrame.Panel)
        self.setFrameShadow(QtWidgets.QFrame.Plain)
        QtWidgets.QTextEdit.focusInEvent(self, event)

    def focusOutEvent(self, event):
        """Specialised handler for focus events.

        Repaint differently the `Logger` frame when it looses the keyboard
        focus so that users can realize easily about this focus change.

        :Parameter event: the event being processed
        """

        self.setLineWidth(self.frame_style['lwidth'])
        self.setFrameShape(self.frame_style['shape'])
        self.setFrameShadow(self.frame_style['shadow'])
        QtWidgets.QTextEdit.focusOutEvent(self, event)
